from gaphor.diagram.drop import drop
from gaphor.diagram.presentation import connect
from gaphor.diagram.support import get_diagram_item
from gaphor.UML.uml import ActivityParameterNode


@drop.register
def drop_relationship(element: ActivityParameterNode, diagram, x, y):
    item_class = get_diagram_item(type(element))
    if not item_class:
        return None

    def drop_distance_to_item(item):
        local_x, local_y = item.matrix_i2c.inverse().transform_point(x, y)
        return item.point(local_x, local_y)

    appendable_item = min(
        diagram.select(
            lambda item: element.activity and element.activity is item.subject
        ),
        key=drop_distance_to_item,
        default=None,
    )

    if not appendable_item:
        return None

    item = diagram.create(item_class, subject=element)
    item.matrix.translate(x, y)
    item.change_parent(appendable_item)
    connect(item, item.handles()[0], appendable_item)

    return item



from typing import Union

from gaphas.connector import Handle, Port

from gaphor import UML
from gaphor.diagram.connectors import Connector, RelationshipConnect
from gaphor.SysML import sysml
from gaphor.SysML.blocks.block import BlockItem
from gaphor.SysML.blocks.interfaceblock import InterfaceBlockItem
from gaphor.SysML.blocks.property import PropertyItem
from gaphor.SysML.blocks.proxyport import ProxyPortItem
from gaphor.UML.deployments import ConnectorItem


@Connector.register(InterfaceBlockItem, ProxyPortItem)
@Connector.register(BlockItem, ProxyPortItem)
@Connector.register(PropertyItem, ProxyPortItem)
class BlockProperyProxyPortConnector:
    def __init__(
        self,
        block_or_property: Union[BlockItem, PropertyItem, InterfaceBlockItem],
        proxy_port: ProxyPortItem,
    ) -> None:
        self.element = block_or_property
        self.proxy_port = proxy_port

    def allow(self, handle: Handle, port: Port) -> bool:
        return (
            bool(self.element.diagram)
            and self.element.diagram is self.proxy_port.diagram
            and (
                isinstance(self.element.subject, UML.EncapsulatedClassifier)
                or isinstance(self.element.subject, UML.Property)
                and isinstance(self.element.subject.type, UML.EncapsulatedClassifier)
            )
        )

    def connect(self, handle: Handle, port: Port) -> bool:
        """Connect and reconnect at model level.

        Returns `True` if a connection is established.
        """
        proxy_port = self.proxy_port
        if not proxy_port.subject:
            proxy_port.subject = proxy_port.model.create(sysml.ProxyPort)

        if isinstance(self.element.subject, UML.EncapsulatedClassifier):
            proxy_port.subject.encapsulatedClassifier = self.element.subject
        elif isinstance(self.element.subject, UML.Property) and isinstance(
            self.element.subject.type, UML.EncapsulatedClassifier
        ):
            proxy_port.subject.encapsulatedClassifier = self.element.subject.type

        # This raises the item in the item hierarchy
        assert proxy_port.diagram
        assert self.element.diagram is proxy_port.diagram
        proxy_port.change_parent(self.element)

        return True

    def disconnect(self, handle: Handle) -> None:
        proxy_port = self.proxy_port
        if proxy_port.subject and proxy_port.diagram:
            proxy_port.change_parent(None)
            proxy_port.subject = None


@Connector.register(ProxyPortItem, ConnectorItem)
@Connector.register(PropertyItem, ConnectorItem)
class PropertyConnectorConnector(RelationshipConnect):
    """Connect a Connector to a Port or Property."""

    line: ConnectorItem
    element: Union[PropertyItem, ProxyPortItem]

    def allow(self, handle, port):
        element = self.element

        # Element should be connected -> have a subject
        return super().allow(handle, port) and isinstance(
            element.subject, (UML.Port, UML.Property)
        )

    def connect_subject(self, handle):
        line = self.line

        c1 = self.get_connected(line.head)
        c2 = self.get_connected(line.tail)
        if c1 and c2 and not line.subject:
            assert isinstance(c1.subject, UML.ConnectableElement)
            assert isinstance(c2.subject, UML.ConnectableElement)
            connector = UML.recipes.create_connector(c1.subject, c2.subject)
            line.subject = connector
            connector.structuredClassifier = c1.subject.owner or c2.subject.owner



from gaphas.tool import item_tool as _item_tool
from gi.repository import Gtk


def item_tool() -> Gtk.GestureDrag:
    gesture = _item_tool()
    gesture.connect_after("drag-end", on_drag_end)
    return gesture


def on_drag_end(gesture, _offset_x, _offset_y):
    view = gesture.get_widget()
    view.selection.dropzone_item = None



import abc

from gaphor.abc import Service


class UIComponent(Service):
    """A user interface component."""

    @abc.abstractmethod
    def open(self):
        """Create and display the UI components (windows)."""

    @abc.abstractmethod
    def close(self):
        """Close the UI component.

        The component can decide to hide or destroy the UI components.
        """

    def shutdown(self):
        """Shut down this component.

        It's not supposed to be opened again.
        """
        self.close()



"""Connector connections.

Implemented using interface item in assembly connector mode, see
`gaphor.diagram.connector` module for details.
"""

from typing import Union

from gaphor import UML
from gaphor.diagram.connectors import BaseConnector, Connector
from gaphor.UML.classes.component import ComponentItem
from gaphor.UML.classes.interface import InterfaceItem
from gaphor.UML.deployments.connector import ConnectorItem


@Connector.register(ComponentItem, ConnectorItem)
@Connector.register(InterfaceItem, ConnectorItem)
class LegacyConnectorConnectBase(BaseConnector):
    """This connector is left as is, mainly for backwards compatibility.

    The Connector item has been removed from the Components tool
    palette.
    """

    element: Union[ComponentItem, InterfaceItem]
    line: ConnectorItem

    def get_connecting(self, iface, both=False):
        """Get items connecting to interface.

        :Parameters:
         iface
            Interface in question.
         both
            If true, then filter out one-side connections.
        """
        diagram = iface.diagram
        connected = diagram.connections.get_connections(connected=iface)
        if both:
            connected = [
                c
                for c in connected
                if diagram.connections.get_connection(c.item.opposite(c.handle))
            ]
        return connected

    @staticmethod
    def get_component(connector):
        """Get component connected by connector."""
        diagram = connector.diagram
        c1 = diagram.connections.get_connection(connector.head)
        c2 = diagram.connections.get_connection(connector.tail)
        component = None
        if c1 and isinstance(c1.connected, ComponentItem):
            component = c1.connected
        elif c2 and isinstance(c2.connected, ComponentItem):
            component = c2.connected
        return component

    def create_uml(self, connector, component, assembly, iface):
        """Create assembly connector UML metamodel for given connector item and
        component.

        :Parameters:
         connector
            Connector item.
         component
            Component item.
         assembly
            Instance of Connector UML metaclass.
         iface
            Instance of Interface UML metaclass.
        """
        connector.subject = assembly

        end = connector.model.create(UML.ConnectorEnd)
        end.role = iface
        end.partWithPort = connector.model.create(UML.Port)
        assembly.end = end

        component.subject.ownedPort = end.partWithPort

    def drop_uml(self, connector, component):
        """Destroy assembly connector UML metamodel existing between connector
        item and component item.

        :Parameters:
         connector
            Connector item.
         component
            Component item.
        """
        if component and component.subject:
            p = component.subject.ownedPort[0]
            p.unlink()
            connector.subject = None

    def allow(self, handle, port):
        iface = self.element
        component = self.get_connected(self.line.opposite(handle))

        if isinstance(component, InterfaceItem) or isinstance(iface, ComponentItem):
            component, iface = iface, component
        # connect only components and interfaces but not two interfaces nor
        # two components
        glue_ok = (not component or isinstance(component, ComponentItem)) and (
            not iface or isinstance(iface, InterfaceItem)
        )

        return glue_ok and super().allow(handle, port)

    def connect(self, handle, port):
        super().connect(handle, port)

        line = self.line

        c1 = self.get_connected(line.head)
        c2 = self.get_connected(line.tail)
        if c1 and c2:
            # reference interface and component correctly
            iface = c1
            component = c2
            if isinstance(component, InterfaceItem):
                assert isinstance(iface, ComponentItem)
                component, iface = iface, component

            connections = self.get_connecting(iface, both=True)

            assembly = None
            for c in connections:
                if c.item.subject:
                    assembly = c.item.subject
                    assert assembly.kind == "assembly"
                    break

            if line.subject:
                assert isinstance(line.subject, UML.Connector)
                assembly = line.subject
                assert assembly.end
                assert list(assembly.end[:].partWithPort)
                assert iface.subject in assembly.end[:].role
                return

            if assembly is None:
                assembly = self.element.model.create(UML.Connector)
                assembly.kind = "assembly"
                for c in connections:
                    connector = c.item
                    self.create_uml(
                        connector,
                        self.get_component(connector),
                        assembly,
                        iface.subject,
                    )
            else:
                self.create_uml(line, component, assembly, iface.subject)
            iface.request_update()

    def disconnect(self, handle):
        super().disconnect(handle)
        line = self.line
        if line.subject is None:
            return

        iface = self.get_connected(line.head)
        if not isinstance(iface, InterfaceItem):
            iface = self.get_connected(line.tail)

        assert iface, "No interface found on {line}"

        connections = list(self.get_connecting(iface, both=True))

        # destroy whole assembly if one connected item stays
        # or only one port will stay connected
        if len(connections) == 2:
            connector = line.subject
            for ci in connections:
                c = self.get_component(ci.item)
                self.drop_uml(ci.item, c)
                line.request_update()
            connector.unlink()
        else:
            c = self.get_component(line)
            self.drop_uml(line, c)
        iface.request_update()



from gi.repository import Gdk, GObject, Gtk

from gaphor.diagram.diagramtoolbox import get_tool_def
from gaphor.diagram.drop import drop
from gaphor.diagram.tools.placement import create_item
from gaphor.event import Notification
from gaphor.i18n import gettext
from gaphor.transaction import Transaction


class ElementDragData(GObject.Object):
    elements = GObject.Property(type=object)


class ToolboxActionDragData(GObject.Object):
    action = GObject.Property(type=str)


def drop_target_tool(modeling_language, event_manager) -> Gtk.EventController:
    """DropTarget tool."""
    drop_target = Gtk.DropTarget.new(GObject.TYPE_NONE, Gdk.DragAction.COPY)
    drop_target.set_gtypes([ElementDragData.__gtype__, ToolboxActionDragData.__gtype__])
    drop_target.connect("drop", on_drop, modeling_language, event_manager)
    return drop_target


def on_drop(target, source_value, x, y, modeling_language, event_manager):
    view = target.get_widget()
    if isinstance(source_value, ElementDragData):
        elements = source_value.elements
        x, y = view.matrix.inverse().transform_point(x, y)
        with Transaction(event_manager):
            items = []
            for element in elements:
                if item := drop(element, view.model, x, y):
                    x += 20
                    y += 20
                    items.append(item)

            if len(items) > 1:
                view.selection.unselect_all()
                view.selection.select_items(*items)
                return True
            if items:
                view.selection.unselect_all()
                view.selection.focused_item = item
                return True
            else:
                event_manager.handle(
                    Notification(gettext("Element can’t be represented on a diagram."))
                )
        return
    elif isinstance(source_value, ToolboxActionDragData):
        tool_def = get_tool_def(modeling_language, source_value.action)
        with Transaction(event_manager):
            create_item(view, tool_def.item_factory, x, y)
        return True

    return False



"""The main application window."""

from __future__ import annotations

import importlib.resources
import logging
from pathlib import Path
from typing import Callable

from gi.repository import Gio, GLib, Gtk

from gaphor.abc import ActionProvider, Service
from gaphor.application import distribution
from gaphor.core import action, event_handler, gettext
from gaphor.event import (
    ActionEnabled,
    ActiveSessionChanged,
    ModelLoaded,
    ModelSaved,
    Notification,
    SessionShutdownRequested,
)
from gaphor.i18n import translated_ui_string
from gaphor.services.modelinglanguage import ModelingLanguageChanged
from gaphor.services.undomanager import UndoManagerStateChanged
from gaphor.ui.abc import UIComponent
from gaphor.ui.actiongroup import window_action_group
from gaphor.ui.event import CurrentDiagramChanged
from gaphor.ui.filedialog import pretty_path
from gaphor.ui.layout import deserialize, is_maximized
from gaphor.ui.modelbrowser import create_diagram_types_model
from gaphor.ui.notification import InAppNotifier

log = logging.getLogger(__name__)


def new_builder():
    builder = Gtk.Builder()
    builder.add_from_string(translated_ui_string("gaphor.ui", "mainwindow.ui"))
    return builder


def create_hamburger_model(export_menu, tools_menu):
    model = Gio.Menu.new()

    part = Gio.Menu.new()
    part.append(gettext("New Model…"), "app.new-model")
    part.append(gettext("Open Model…"), "app.file-open")
    model.append_section(None, part)

    part = Gio.Menu.new()
    part.append(gettext("Save"), "win.file-save")
    part.append(gettext("Save As…"), "win.file-save-as")
    part.append_submenu(gettext("Export"), export_menu)
    model.append_section(None, part)

    part = Gio.Menu.new()
    part.append_submenu(gettext("Tools"), tools_menu)
    model.append_section(None, part)

    part = Gio.Menu.new()
    part.append(gettext("Preferences"), "app.preferences")
    part.append(gettext("Keyboard Shortcuts"), "app.shortcuts")
    part.append(gettext("About Gaphor"), "app.about")
    model.append_section(None, part)

    return model


def create_modeling_language_model(modeling_language):
    model = Gio.Menu.new()
    for id, name in modeling_language.modeling_languages:
        menu_item = Gio.MenuItem.new(name, "win.select-modeling-language")
        menu_item.set_attribute_value("target", GLib.Variant.new_string(id))
        model.append_item(menu_item)
    return model


class MainWindow(Service, ActionProvider):
    """The main window for the application.

    It contains a Namespace-based tree view and a menu and a statusbar.
    """

    def __init__(
        self,
        event_manager,
        component_registry,
        properties,
        modeling_language,
        export_menu,
        tools_menu,
    ):
        self.event_manager = event_manager
        self.component_registry = component_registry
        self.properties = properties
        self.modeling_language = modeling_language
        self.export_menu = export_menu
        self.tools_menu = tools_menu

        self._builder: Gtk.Builder | None = new_builder()
        self.action_group: Gio.ActionGroup = None
        self.modeling_language_name = None
        self.diagram_types = None
        self.in_app_notifier = None

        # UI updates to be performed when the window is shown/open.
        self._ui_updates: list[Callable[[], None]] = []

        event_manager.subscribe(self._on_file_manager_state_changed)
        event_manager.subscribe(self._on_current_diagram_changed)

    def shutdown(self):
        if self.window:
            self.window.destroy()
            self._builder = None

        em = self.event_manager
        em.unsubscribe(self._on_file_manager_state_changed)
        em.unsubscribe(self._on_current_diagram_changed)
        em.unsubscribe(self._on_undo_manager_state_changed)
        em.unsubscribe(self._on_action_enabled)
        em.unsubscribe(self._on_modeling_language_selection_changed)
        if self.in_app_notifier:
            em.unsubscribe(self.in_app_notifier.handle)
            self.in_app_notifier = None

    @property
    def window(self):
        return self._builder.get_object("main-window") if self._builder else None

    @property
    def title(self):
        return self._builder.get_object("title") if self._builder else None

    @property
    def modified(self):
        return self._builder.get_object("modified") if self._builder else None

    @property
    def subtitle(self):
        return self._builder.get_object("subtitle") if self._builder else None

    @property
    def model_changed(self) -> bool:
        return self.modified.get_visible() if self.modified else False  # type: ignore[no-any-return]

    @model_changed.setter
    def model_changed(self, model_changed: bool):
        if self.modified:
            self.modified.set_visible(model_changed)

    def get_ui_component(self, name):
        return self.component_registry.get(UIComponent, name)

    def open(self, gtk_app=None):
        """Open the main window."""

        builder = self._builder
        assert builder
        window = self.window
        window.set_application(gtk_app)
        if ".dev" in distribution().version:
            window.get_style_context().add_class("devel")

        select_modeling_language = builder.get_object("select-modeling-language")
        select_modeling_language.set_menu_model(
            create_modeling_language_model(self.modeling_language),
        )
        self.modeling_language_name = builder.get_object("modeling-language-name")

        self.diagram_types = builder.get_object("diagram-types")
        self.diagram_types.set_menu_model(
            create_diagram_types_model(self.modeling_language)
        )

        hamburger = builder.get_object("hamburger")
        hamburger.set_menu_model(
            create_hamburger_model(self.export_menu.menu, self.tools_menu.menu),
        )

        window.set_default_size(*(self.properties.get("ui.window-size", (860, 580))))
        if self.properties.get("ui.window-mode", "") == "maximized":
            window.maximize()
        elif self.properties.get("ui.window-mode", "") == "fullscreened":
            window.fullscreen()

        def _factory(name):
            comp = self.get_ui_component(name)
            return comp.open()

        main_content = builder.get_object("main-content")
        deserialize(
            main_content,
            (importlib.resources.files("gaphor.ui") / "layout.xml").read_text(
                encoding="utf-8"
            ),
            _factory,
            self.properties,
        )

        self.action_group, shortcuts = window_action_group(self.component_registry)
        window.insert_action_group("win", self.action_group)

        self._on_modeling_language_selection_changed()

        window.set_resizable(True)
        window.add_controller(Gtk.ShortcutController.new_for_model(shortcuts))
        window.connect("close-request", self._on_window_close_request)
        window.connect("notify::default-height", self._on_window_size_changed)
        window.connect("notify::default-width", self._on_window_size_changed)
        window.connect("notify::maximized", self._on_window_mode_changed)
        window.connect("notify::fullscreened", self._on_window_mode_changed)
        window.set_visible(True)

        window.connect("notify::is-active", self._on_window_active)

        self.in_app_notifier = InAppNotifier(main_content)
        em = self.event_manager
        em.subscribe(self._on_undo_manager_state_changed)
        em.subscribe(self._on_action_enabled)
        em.subscribe(self._on_modeling_language_selection_changed)
        em.subscribe(self.in_app_notifier.handle)

        for handler in self._ui_updates:
            handler()
        del self._ui_updates[:]

    # Actions:

    @action("fullscreen", shortcut="F11", state=False)
    def toggle_fullscreen(self, active):
        if not self.window:
            return

        if active:
            self.window.fullscreen()
            self.event_manager.handle(
                Notification(gettext("Press F11 to exit Fullscreen"))
            )
        else:
            self.window.unfullscreen()

    # Signal callbacks:

    @event_handler(ModelLoaded, ModelSaved)
    def _on_file_manager_state_changed(self, event: ModelLoaded | ModelSaved) -> None:
        if not (window := self.window):
            self._ui_updates.append(lambda: self._on_file_manager_state_changed(event))
            return

        filename = Path(event.filename) if event.filename else None

        self.subtitle.set_text(
            pretty_path(filename) if filename else gettext("New model")
        )
        window.set_title(
            f"{filename.name} ({pretty_path(filename.parent)}) - Gaphor"
            if filename
            else f"{gettext('New model')} - Gaphor"
        )

        self.model_changed = isinstance(event, ModelLoaded) and event.modified

        window.present()

    @event_handler(CurrentDiagramChanged)
    def _on_current_diagram_changed(self, event):
        if not self.window:
            self._ui_updates.append(lambda: self._on_current_diagram_changed(event))
            return

        self.title.set_text(
            (event.diagram.name or gettext("<None>")) if event.diagram else "Gaphor"
        )

    @event_handler(UndoManagerStateChanged)
    def _on_undo_manager_state_changed(self, event):
        undo_manager = event.service
        self.model_changed = undo_manager.can_undo()

    @event_handler(ActionEnabled)
    def _on_action_enabled(self, event):
        if self.action_group and event.scope == "win":
            a = self.action_group.lookup_action(event.name)
            a.set_enabled(event.enabled)

    @event_handler(ModelingLanguageChanged)
    def _on_modeling_language_selection_changed(self, event=None):
        if self.modeling_language_name:
            self.modeling_language_name.set_label(
                gettext("Profile: {name}").format(name=self.modeling_language.name)
            )
        if self.diagram_types:
            self.diagram_types.set_menu_model(
                create_diagram_types_model(self.modeling_language)
            )

    def _on_window_active(self, window, prop):
        self.event_manager.handle(ActiveSessionChanged(self))

    def _on_window_close_request(self, window, event=None):
        self.event_manager.handle(SessionShutdownRequested(self))
        return True

    def _on_window_size_changed(self, window, _gspec):
        if not is_maximized(window):
            width, height = window.get_default_size()
            self.properties.set("ui.window-size", (width, height))

    def _on_window_mode_changed(self, window, gspec):
        mode = gspec.name
        self.properties.set("ui.window-mode", mode if window.get_property(mode) else "")



from __future__ import annotations

import tinycss2.color3
from gi.repository import Gio, GObject, Gtk, GtkSource

from gaphor.core.styling import Style


class ProposalBase:
    def proposals(self) -> Gio.ListStore:
        raise NotImplementedError()

    def proposal_text(self, proposal: GtkSource.CompletionProposal) -> str:
        raise NotImplementedError()

    def position_cursor(self, buffer, cursor):
        pass


class CssNamedColorProposals(ProposalBase):
    def __init__(self):
        super().__init__()
        store = Gio.ListStore.new(TextProposal)
        for color_name in tinycss2.color3._COLOR_KEYWORDS:
            proposal = TextProposal(color_name)
            store.append(proposal)
        self.store = store

    def proposals(self):
        return self.store

    def proposal_text(self, proposal: GtkSource.CompletionProposal) -> str:
        return proposal.text  # type: ignore[no-any-return]


class CssFunctionProposals(ProposalBase):
    FUNCTIONS = [
        "hsl",
        "hsla",
        "rgb",
        "rgba",
    ]

    def __init__(self):
        super().__init__()
        store = Gio.ListStore.new(TextProposal)
        for function in self.FUNCTIONS:
            proposal = TextProposal(function)
            store.append(proposal)
        self.store = store

    def proposals(self):
        return self.store

    def proposal_text(self, proposal: GtkSource.CompletionProposal) -> str:
        return f"{proposal.text}()"

    def position_cursor(self, buffer, cursor):
        cursor.backward_char()
        buffer.place_cursor(cursor)


class CssPropertyProposals(ProposalBase):
    PROPERTIES = sorted(Style.__optional_keys__)

    def __init__(self):
        super().__init__()

        store = Gio.ListStore.new(TextProposal)
        for prop in self.PROPERTIES:
            proposal = TextProposal(prop)
            store.append(proposal)
        self.store = store

    def proposals(self):
        return self.store

    def proposal_text(self, proposal):
        return f"{proposal.text}: "


class CompletionProviderWrapper(GObject.GObject, GtkSource.CompletionProvider):
    def __init__(self, priority: int, proposals: ProposalBase):
        super().__init__()
        self._priority = priority
        self._proposals = proposals
        self._filter_data: FilterData = FilterData()

    def do_activate(
        self,
        context: GtkSource.CompletionContext,
        proposal: GtkSource.CompletionProposal,
    ) -> None:
        buffer = context.get_buffer()
        buffer.begin_user_action()
        has_selection, begin, end = context.get_bounds()
        if has_selection:
            buffer.delete(begin, end)
        text = self._proposals.proposal_text(proposal)
        buffer.insert(begin, text, len(text))
        self._proposals.position_cursor(buffer, begin)
        buffer.end_user_action()

    def do_display(
        self,
        context: GtkSource.CompletionContext,
        proposal: GtkSource.CompletionProposal,
        cell: GtkSource.CompletionCell,
    ) -> None:
        if cell.props.column == GtkSource.CompletionColumn.ICON:
            pass
        elif cell.props.column == GtkSource.CompletionColumn.TYPED_TEXT:
            cell.set_text(proposal.text)

    def do_get_priority(self, context: GtkSource.CompletionContext) -> int:
        return self._priority

    def do_get_title(self) -> str:
        return self.__class__.__name__

    def do_populate_async(self, context, cancellable, callback, user_data=None) -> None:
        self._filter_data.word = context.get_word()
        store = self._proposals.proposals()

        def filter_fn(proposal, data):
            return proposal.text.startswith(data.word)

        store_filter = Gtk.CustomFilter.new(filter_fn, self._filter_data)
        proposals = Gtk.FilterListModel.new(store, store_filter)
        context.set_proposals_for_provider(self, proposals)

    def do_refilter(
        self, context: GtkSource.CompletionContext, model: Gio.ListModel
    ) -> None:
        word = context.get_word()
        change = Gtk.FilterChange.DIFFERENT
        if old_word := self._filter_data.word:
            if word.startswith(old_word):
                change = Gtk.FilterChange.MORE_STRICT
            elif old_word.startswith(word):
                change = Gtk.FilterChange.LESS_STRICT
        self._filter_data.word = word
        model.get_filter().changed(change)


class TextProposal(GObject.Object, GtkSource.CompletionProposal):
    def __init__(self, text: str):
        super().__init__()
        self.text: str = text


class FilterData:
    word: str



from __future__ import annotations

from typing import Iterator, NamedTuple

from gaphor.core.modeling.element import Id
from gaphor.diagram.copypaste import (
    ElementCopy,
    copy,
    copy_element,
    paste,
    paste_element,
)
from gaphor.UML.recipes import owner_package
from gaphor.UML.uml import NamedElement, Package, Relationship, Type


class NamedElementCopy(NamedTuple):
    element_copy: ElementCopy


def copy_named_element(
    element: NamedElement, blacklist: list[str] | None = None
) -> NamedElementCopy:
    return NamedElementCopy(
        element_copy=copy_element(element, blacklist),
    )


@copy.register
def _copy_named_element(element: NamedElement) -> Iterator[tuple[Id, NamedElementCopy]]:
    yield element.id, copy_named_element(element)


def paste_named_element(copy_data: NamedElementCopy, diagram, lookup):
    paster = paste_element(
        copy_data.element_copy,
        diagram,
        lookup,
        filter=lambda n, v: not isinstance(v, Relationship),
    )
    element = next(paster)
    yield element
    next(paster, None)
    if (
        not element.namespace
        and isinstance(element, (Type, Package))
        and (package := owner_package(diagram.owner)) is not element
    ):
        element.package = package


paste.register(NamedElementCopy, paste_named_element)



"""Copy and deep/shallow paste elements in a model or between models.

The `copy_full()` function will return all values serialized, either as string
values or reference id's.

The `paste_link()` function will resolve those values and place instances of
them on the diagram using the same defining element.
The `paste_full()` function is similar, but it creates new defining elements
for the elements being pasted.

The `copy()` dispatch function is an iterator that returns a copy of the element
and optionally any related elements. E.g. a Presentation element is likely
to also create a copy of its subject.

The `paste()` dispatch function works in two parts. First it creates the element
and yields it. On the next invocation it will populate the element and perform
complete the model loading.

`copy()` and `paste()` use `Element`'s `save()` and `load()` methods.
"""

from __future__ import annotations

from collections.abc import Iterable
from functools import singledispatch
from typing import Callable, Collection, Iterator, NamedTuple

from gaphor.core.modeling import Diagram, Presentation
from gaphor.core.modeling.collection import collection
from gaphor.core.modeling.element import Element, Id

Opaque = object


class CopyData(NamedTuple):
    elements: dict[Id, Opaque]
    diagram_refs: set[Id]


def copy_full(
    items: Collection, lookup: Callable[[Id], Element | None] | None = None
) -> CopyData:
    """Copy items, including owned elements."""
    elements = {ref: data for item in items for ref, data in copy(item)}
    diagram_refs = {item.diagram.id for item in items if isinstance(item, Presentation)}

    if not lookup:
        return CopyData(elements=elements, diagram_refs=diagram_refs)

    def copy_owned(e):
        for o in e.ownedElement:
            if o.owner is e:
                for ref, data in copy(o):
                    if ref not in elements:
                        elements[ref] = data
                        copy_owned(o)

    for ref in list(elements.keys()):
        copy_owned(lookup(ref))

    return CopyData(elements=elements, diagram_refs=diagram_refs)


def paste_link(copy_data: Opaque, diagram: Diagram) -> set[Presentation]:
    return _paste(copy_data, diagram, full=False)


def paste_full(copy_data: Opaque, diagram: Diagram) -> set[Presentation]:
    return _paste(copy_data, diagram, full=True)


@singledispatch
def copy(obj: Element | Iterable) -> Iterator[tuple[Id, Opaque]]:
    """Create a copy of an element (or list of elements).

    The returned type should be distinct, so the `paste()` function can
    properly dispatch.
    """
    raise ValueError(f"No copier for {obj}")


@singledispatch
def paste(
    copy_data: Opaque, diagram: Diagram, lookup: Callable[[Id], Element | None]
) -> Iterator[Element]:
    """Paste previously copied data.

    Based on the data type created in the `copy()` function, try to
    duplicate the copied elements.
    """
    raise ValueError(f"No paster for {copy_data}")


def serialize(value):
    if isinstance(value, Element):
        return ("r", value.id)
    elif isinstance(value, collection):
        return ("c", [serialize(v) for v in value])
    else:
        if isinstance(value, bool):
            value = int(value)
        return ("v", str(value))


def deserialize(ser, lookup):
    vtype, value = ser
    if vtype == "r":
        if e := lookup(value):
            yield e
    elif vtype == "c":
        for v in value:
            yield from deserialize(v, lookup)
    elif vtype == "v":
        yield value


class ElementCopy(NamedTuple):
    cls: type[Element]
    id: Id
    data: dict[str, tuple[str, str]]


def copy_element(element: Element, blacklist: list[str] | None = None) -> ElementCopy:
    data = {}
    # do not copy Element.presentation, to avoid cyclic dependencies
    blacklist_ = blacklist + ["presentation"] if blacklist else ["presentation"]

    def save_func(name, value):
        if name not in blacklist_:
            data[name] = serialize(value)

    element.save(save_func)
    return ElementCopy(cls=element.__class__, id=element.id, data=data)


@copy.register
def _copy_element(element: Element) -> Iterator[tuple[Id, ElementCopy]]:
    yield element.id, copy_element(element)


def paste_element(
    copy_data: ElementCopy,
    diagram,
    lookup,
    filter: Callable[[str, str | int | Element], bool] | None = None,
):
    cls, _id, data = copy_data
    element = diagram.model.create(cls)
    yield element
    for name, ser in data.items():
        for value in deserialize(ser, lookup):
            if not filter or filter(name, value):
                element.load(name, value)
    element.postload()


paste.register(ElementCopy, paste_element)


@copy.register
def _copy_diagram(element: Diagram) -> Iterator[tuple[Id, ElementCopy]]:
    yield element.id, copy_element(element, blacklist=["ownedPresentation"])


class PresentationCopy(NamedTuple):
    cls: type[Element]
    data: dict[str, tuple[str, str]]
    diagram: Id
    parent: Id | None


def copy_presentation(item: Presentation) -> PresentationCopy:
    assert item.diagram

    parent = item.parent
    return PresentationCopy(
        cls=item.__class__,
        data=copy_element(item, blacklist=["diagram", "parent", "children"]).data,
        diagram=item.diagram.id,
        parent=parent.id if parent else None,
    )


@copy.register
def _copy_presentation(item: Presentation) -> Iterator[tuple[Id, object]]:
    yield item.id, copy_presentation(item)
    if item.subject:
        yield from copy(item.subject)


@paste.register
def _paste_presentation(copy_data: PresentationCopy, _diagram, lookup):
    cls, data, diagram_ref, parent = copy_data
    diagram = lookup(diagram_ref)
    item = diagram.create(cls)
    yield item
    if parent:
        if p := lookup(parent):
            item.parent = p

    for name, ser in data.items():
        for value in deserialize(ser, lookup):
            item.load(name, value)
    diagram.update_now((item,))


def _paste(copy_data: Opaque, diagram: Diagram, full: bool) -> set[Presentation]:
    assert isinstance(copy_data, CopyData)

    model = diagram.model

    # Map the original diagram ids to our new target diagram
    new_elements: dict[Id, Element] = {ref: diagram for ref in copy_data.diagram_refs}

    def element_lookup(ref: Id):
        if ref in new_elements:
            return new_elements[ref]

        looked_up = model.lookup(ref)
        if not full and looked_up and not isinstance(looked_up, Presentation):
            return looked_up

        if ref in copy_data.elements:
            paster = paste(copy_data.elements[ref], diagram, element_lookup)
            new_elements[ref] = next(paster)
            next(paster, None)
            return new_elements[ref]

        if full and looked_up and not isinstance(looked_up, Presentation):
            return looked_up

        if looked_up := diagram.lookup(ref):
            return looked_up

    for old_id in copy_data.elements.keys():
        if old_id in new_elements:
            continue
        element_lookup(old_id)

    for element in new_elements.values():
        assert element
        element.postload()

    return {
        e
        for e in new_elements.values()
        if isinstance(e, Presentation) and e.diagram is diagram
    }



"""Tools for handling items on a diagram."""

# ruff: noqa: F401

from gaphas.tool import hover_tool, rubberband_tool, view_focus_tool, zoom_tools
from gaphas.tool.scroll import pan_tool

import gaphor.diagram.tools.handlemove
import gaphor.diagram.tools.segment
from gaphor.diagram.tools.dnd import drop_target_tool
from gaphor.diagram.tools.dropzone import drop_zone_tool
from gaphor.diagram.tools.itemtool import item_tool
from gaphor.diagram.tools.magnet import magnet_tool
from gaphor.diagram.tools.placement import placement_tool
from gaphor.diagram.tools.shortcut import shortcut_tool
from gaphor.diagram.tools.textedit import text_edit_tools
from gaphor.diagram.tools.txtool import transactional_tool


def apply_default_tool_set(view, modeling_language, event_manager, rubberband_state):
    """The default tool set."""
    view.remove_all_controllers()
    view.add_controller(hover_tool())
    view.add_controller(*text_edit_tools(event_manager))
    view.add_controller(*transactional_tool(item_tool(), event_manager=event_manager))
    view.add_controller(rubberband_tool(rubberband_state))
    add_basic_tools(view, modeling_language, event_manager)


def apply_magnet_tool_set(view, modeling_language, event_manager):
    """The default tool set."""
    view.remove_all_controllers()
    view.add_controller(
        *transactional_tool(magnet_tool(event_manager), event_manager=event_manager)
    )
    add_basic_tools(view, modeling_language, event_manager)


def apply_placement_tool_set(
    view, item_factory, modeling_language, event_manager, handle_index
):
    view.remove_all_controllers()
    view.add_controller(view_focus_tool())
    view.add_controller(
        *transactional_tool(
            placement_tool(item_factory, event_manager, handle_index),
            event_manager=event_manager,
        )
    )
    view.add_controller(
        drop_zone_tool(item_factory.item_class, item_factory.subject_class)
    )
    add_basic_tools(view, modeling_language, event_manager)


def add_basic_tools(view, modeling_language, event_manager):
    for tool in zoom_tools():
        view.add_controller(tool)
    view.add_controller(pan_tool())
    view.add_controller(view_focus_tool())
    view.add_controller(shortcut_tool(event_manager))
    view.add_controller(drop_target_tool(modeling_language, event_manager))



"""Majority vote gate item definition."""

from math import pi

from gaphas.geometry import Rectangle

from gaphor.core.modeling import DrawContext
from gaphor.diagram.presentation import (
    Classified,
    ElementPresentation,
    from_package_str,
    text_name,
)
from gaphor.diagram.shapes import Box, IconBox, Text, stroke
from gaphor.diagram.support import represents
from gaphor.RAAML import raaml
from gaphor.RAAML.fta.constants import DEFAULT_FTA_MAJOR, DEFAULT_FTA_MINOR
from gaphor.UML.recipes import stereotypes_str

DEFAULT_WIDTH = 40


@represents(raaml.MAJORITY_VOTE)
class MajorityVoteItem(Classified, ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=DEFAULT_WIDTH, height=DEFAULT_FTA_MAJOR)

        self.watch("subject[NamedElement].name").watch(
            "subject[NamedElement].namespace.name"
        )

    def update_shapes(self, event=None):
        self.shape = IconBox(
            Box(
                draw=draw_majority_vote_gate,
            ),
            Text(
                text=lambda: stereotypes_str(
                    self.subject, [self.diagram.gettext("Majority Vote Gate")]
                ),
            ),
            text_name(self),
            Text(
                text=lambda: from_package_str(self),
                style={"font-size": "x-small"},
            ),
        )


def draw_majority_vote_gate(box, context: DrawContext, bounding_box: Rectangle):
    cr = context.cairo
    left = 0
    right = bounding_box.width
    wall_top = bounding_box.height / 3.0
    wall_bottom = bounding_box.height

    # Left wall
    cr.move_to(left, wall_bottom)
    cr.line_to(left, wall_top)

    # Right wall
    cr.move_to(right, wall_bottom)
    cr.line_to(right, wall_top)

    # Top arc
    rx = right - left
    ry = bounding_box.height * 2.0 / 3.0
    cr.move_to(left, wall_top)
    cr.save()
    cr.translate(left + rx / 2.0, wall_top)
    cr.scale(rx / 2.0, ry / 2.0)
    cr.arc(0.0, 0.0, 1.0, pi, 0)
    cr.restore()

    # Bottom arc
    ry = bounding_box.height / 4.0
    cr.move_to(left, wall_bottom)
    cr.save()
    cr.translate(left + rx / 2.0, wall_bottom)
    cr.scale(rx / 2.0, ry / 2.0)
    cr.arc(0.0, 0.0, 1.0, pi, 0)
    cr.restore()

    # Draw "m"
    text = "m"
    if bounding_box.height > 3 * bounding_box.width:
        cr.set_font_size(32 * bounding_box.width / DEFAULT_WIDTH)
    elif bounding_box.width > 3 * bounding_box.height:
        cr.set_font_size(40 * bounding_box.height / DEFAULT_FTA_MAJOR)
    else:
        cr.set_font_size(
            17
            * (bounding_box.width + bounding_box.height)
            / (DEFAULT_FTA_MINOR + DEFAULT_FTA_MAJOR)
        )
    x_bearing, y_bearing, width, height, x_advance, y_advance = cr.text_extents(text)
    x = bounding_box.width / 2.0 - (width / 2 + x_bearing)
    y = bounding_box.height / 2.0 - (height / 2 + y_bearing)
    cr.move_to(x, y)
    cr.show_text(text)

    stroke(context, fill=True)



"""The definition for the use cases section of the toolbox."""

from gaphas.item import SE

from gaphor import UML
from gaphor.core import gettext
from gaphor.diagram.diagramtoolbox import ToolDef, ToolSection, new_item_factory
from gaphor.UML import diagramitems
from gaphor.UML.toolboxconfig import namespace_config

use_cases = ToolSection(
    gettext("Use Cases"),
    (
        ToolDef(
            "toolbox-use-case",
            gettext("Use case"),
            "gaphor-use-case-symbolic",
            "u",
            new_item_factory(
                diagramitems.UseCaseItem,
                UML.UseCase,
                config_func=namespace_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-actor",
            gettext("Actor"),
            "gaphor-actor-symbolic",
            "t",
            new_item_factory(
                diagramitems.ActorItem,
                UML.Actor,
                config_func=namespace_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-use-case-association",
            gettext("Association"),
            "gaphor-association-symbolic",
            "<Shift>J",
            new_item_factory(diagramitems.AssociationItem),
        ),
        ToolDef(
            "toolbox-include",
            gettext("Include"),
            "gaphor-include-symbolic",
            "<Shift>U",
            new_item_factory(diagramitems.IncludeItem),
            handle_index=0,
        ),
        ToolDef(
            "toolbox-extend",
            gettext("Extend"),
            "gaphor-extend-symbolic",
            "<Shift>X",
            new_item_factory(diagramitems.ExtendItem),
            handle_index=0,
        ),
    ),
)



"""Stereotype property page."""
from unicodedata import normalize

from gi.repository import Gdk, Gio, GLib, GObject, Gtk

from gaphor import UML
from gaphor.core import transactional
from gaphor.core.modeling.element import Element
from gaphor.diagram.propertypages import PropertyPageBase, PropertyPages
from gaphor.i18n import gettext, translated_ui_string
from gaphor.UML.profiles.metaclasspropertypage import new_builder
from gaphor.UML.propertypages import text_field_handlers


@PropertyPages.register(UML.Element)
class StereotypePage(PropertyPageBase):
    order = 40

    def __init__(self, item):
        self.item = item

    def construct(self):
        subject = self.item.subject
        if not subject:
            return None

        stereotypes = UML.recipes.get_stereotypes(subject)
        if not stereotypes:
            return None

        builder = new_builder(
            "stereotypes-editor",
            signals={
                "show-stereotypes-changed": (self._on_show_stereotypes_change,),
                "stereotype-activated": (stereotype_activated,),
                "stereotype-key-pressed": (stereotype_key_handler,),
            },
        )

        show_stereotypes = builder.get_object("show-stereotypes")

        if hasattr(self.item, "show_stereotypes"):
            show_stereotypes.set_active(self.item.show_stereotypes)
        else:
            show_stereotypes.get_parent().unparent()

        stereotype_list = builder.get_object("stereotype-list")
        model = stereotype_model(subject)

        stereotype_set_model_with_interaction(stereotype_list, model)

        return builder.get_object("stereotypes-editor")

    @transactional
    def _on_show_stereotypes_change(self, button, gparam):
        self.item.show_stereotypes = button.get_active()


def stereotype_set_model_with_interaction(stereotype_list, model):
    def on_stereotype_toggled(cell):
        button = cell.get_child().get_first_child()
        view = cell.get_item()
        view.applied = button.get_active()
        update_stereotype_model(model)

    for column, factory in zip(
        stereotype_list.get_columns(),
        [
            name_list_item_factory(
                signal_handlers={
                    "on_toggled": (on_stereotype_toggled,),
                },
            ),
            value_list_item_factory(
                signal_handlers=text_field_handlers("slot_value"),
            ),
        ],
    ):
        column.set_factory(factory)

    selection = Gtk.SingleSelection.new(model)
    stereotype_list.set_model(selection)


class StereotypeView(GObject.Object):
    def __init__(
        self,
        target: Element,
        stereotype: UML.Stereotype,
        attr: UML.Property | None,
    ):
        super().__init__()
        self.stereotype = stereotype
        self.target = target
        self.attr = attr

    @property
    def instance(self):
        return stereotype_application(self.stereotype, self.target)

    @property
    def slot(self):
        instance = self.instance
        attr = self.attr
        return (
            next(
                (slot for slot in instance.slot if slot.definingFeature is attr),
                None,
            )
            if instance
            else None
        )

    editing = GObject.Property(type=bool, default=False)

    @GObject.Property(type=str, default="", flags=GObject.ParamFlags.READABLE)
    def name(self):
        return (self.attr.name if self.attr else self.stereotype.name) or ""

    @GObject.Property(type=bool, default=False)
    def applied(self):
        return bool(self.instance)

    @applied.setter  # type: ignore[no-redef]
    @transactional
    def applied(self, value):
        instance = self.instance
        if value and not instance:
            UML.recipes.apply_stereotype(self.target, self.stereotype)
        elif not value and instance:
            UML.recipes.remove_stereotype(self.target, self.stereotype)

    @GObject.Property(type=bool, default=False)
    def checkbox_visible(self):
        return not bool(self.attr)

    @GObject.Property(type=str)
    def placeholder_text(self):
        return gettext("New Slot Value…") if self.attr and self.instance else ""

    @GObject.Property(type=bool, default=False)
    def can_edit(self):
        return bool(self.attr and self.instance)

    @GObject.Property(type=GObject.TYPE_STRV)
    def indent_classes(self):
        return ["stereotype-value"] if self.attr else []

    @GObject.Property(type=bool, default=True)
    def sensitive(self):
        return not self.attr or self.instance

    @GObject.Property(type=str)
    def slot_value(self):
        slot = self.slot
        return slot.value if slot else ""

    @slot_value.setter  # type: ignore[no-redef]
    @transactional
    def slot_value(self, value):
        slot = self.slot
        if value:
            if slot is None:
                slot = UML.recipes.add_slot(self.instance, self.attr)
            slot.value = value
        elif slot:
            del self.instance.slot[slot]

    def start_editing(self):
        self.editing = True


def stereotype_model(subject: Element):
    model = Gio.ListStore.new(StereotypeView.__gtype__)
    stereotypes = UML.recipes.get_stereotypes(subject)

    for st in sorted(
        stereotypes, key=lambda s: normalize("NFC", s.name or "").casefold()
    ):
        model.append(StereotypeView(subject, st, None))

        for attr in all_attributes(st):
            model.append(StereotypeView(subject, st, attr))

    return model


def update_stereotype_model(store: Gio.ListStore):
    for view in store:
        view.notify("placeholder_text")
        view.notify("can_edit")
        view.notify("slot_value")
        view.notify("sensitive")


def all_attributes(stereotype, seen=None):
    if seen is None:
        seen = {stereotype}

    for super_type in stereotype.generalization[:].general[:]:
        if super_type not in seen:
            seen.add(super_type)
            yield from all_attributes(super_type)
    yield from (attr for attr in stereotype.ownedAttribute if not attr.association)


def stereotype_application(st, subject) -> UML.InstanceSpecification | None:
    return next(
        (applied for applied in subject.appliedStereotype if st in applied.classifier),
        None,
    )


def name_list_item_factory(signal_handlers=None):
    ui_string = translated_ui_string("gaphor.UML.profiles", "stereotype-name-cell.ui")

    return Gtk.BuilderListItemFactory.new_from_bytes(
        Gtk.Builder.BuilderScope(signal_handlers),
        GLib.Bytes.new(ui_string.encode("utf-8")),
    )


def value_list_item_factory(signal_handlers=None):
    ui_string = translated_ui_string("gaphor.UML.profiles", "stereotype-value-cell.ui")

    return Gtk.BuilderListItemFactory.new_from_bytes(
        Gtk.Builder.BuilderScope(signal_handlers),
        GLib.Bytes.new(ui_string.encode("utf-8")),
    )


@transactional
def stereotype_activated(list_view, _row):
    selection = list_view.get_model()
    item = selection.get_selected_item()

    if item.attr:
        item.start_editing()
    else:
        item.applied = not item.applied


@transactional
def stereotype_key_handler(ctrl, keyval, _keycode, state):
    list_view = ctrl.get_widget()
    selection = list_view.get_model()
    item = selection.get_selected_item()

    if keyval in (Gdk.KEY_F2,):
        item.start_editing()
        return True

    if keyval in (Gdk.KEY_Delete, Gdk.KEY_BackSpace) and not state & (
        Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.SHIFT_MASK
    ):
        item.slot_value = ""
        return True

    return False



from gaphor.core import gettext
from gaphor.diagram.diagramtoolbox import ToolDef, ToolSection, new_item_factory
from gaphor.SysML.allocations.relationships import AllocateItem

allocations = ToolSection(
    gettext("Allocations"),
    (
        ToolDef(
            "toolbox-allocate-dependency",
            gettext("Allocate"),
            "gaphor-allocate-symbolic",
            None,
            new_item_factory(AllocateItem),
        ),
    ),
)



from gi.repository import Gdk, GLib, GObject, Gtk, Pango

START_EDIT_DELAY = 100  # ms

_XML = """\
<?xml version="1.0" encoding="UTF-8"?>
<interface>
  <requires lib="gtk" version="4.0"/>
  <template class="TextField" parent="GtkStack">
    <child>
      <object class="GtkStackPage">
        <property name="name">default</property>
        <property name="child">
          <object class="GtkLabel">
            <property name="xalign">0</property>
            <property name="hexpand">yes</property>
            <binding name="attributes">
              <lookup name="attributes">TextField</lookup>
            </binding>
            <binding name="label">
              <lookup name="readonly-text">TextField</lookup>
            </binding>
          </object>
        </property>
      </object>
    </child>
    <child>
      <object class="GtkStackPage">
        <property name="name">editing</property>
        <property name="child">
          <object class="GtkText" id="_text">
            <property name="propagate-text-width">yes</property>
            <property name="buffer">
              <object class="GtkEntryBuffer">
                <binding name="text">
                  <lookup name="editable-text">TextField</lookup>
                </binding>
              </object>
            </property>
            <child>
              <object class="GtkEventControllerFocus">
                <signal name="leave" handler="on_text_focus_out" />
              </object>
            </child>
            <child>
              <object class="GtkEventControllerKey">
                <signal name="key-pressed" handler="on_text_key_pressed" />
              </object>
            </child>
          </object>
        </property>
      </object>
    </child>
    <child>
      <object class="GtkStackPage">
        <property name="name">placeholder</property>
        <property name="child">
          <object class="GtkLabel">
            <property name="xalign">0</property>
            <property name="hexpand">yes</property>
            <binding name="label">
              <lookup name="placeholder-text">TextField</lookup>
            </binding>
            <style>
              <class name="dim-label" />
            </style>
          </object>
        </property>
      </object>
    </child>
  </template>
</interface>
"""


@Gtk.Template(string=_XML)
class TextField(Gtk.Stack):
    """A text input field for use in ``ListView``s.

    The Text Field contains two layers of text. One for displaying
    (a `Gtk.Label`), and one for editing (`Gtk.Text`).
    This allows us to easily define a separate text for editing.
    For example: a class attribute is represented as `+ attr: int = 0`,
    but if we edit it, we should only be able to edit it's name (`attr`).
    """

    __gtype_name__ = "TextField"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._readonly_text = ""

    editable_text = GObject.Property(type=str, default="")
    placeholder_text = GObject.Property(type=str, default="")
    attributes = GObject.Property(type=Pango.AttrList)
    """Pango style attributes for the readonly text."""

    can_edit = GObject.Property(type=bool, default=True)

    @GObject.Property(type=str, default="")
    def readonly_text(self):
        return self._readonly_text

    @readonly_text.setter  # type: ignore[no-redef]
    def readonly_text(self, value):
        self._readonly_text = value or ""
        self.set_visible_child_name("default" if value else "placeholder")

    @GObject.Property(type=bool, default=False)
    def editing(self):
        return self.get_visible_child_name() == "editing"

    @editing.setter  # type: ignore[no-redef]
    def editing(self, value):
        if not self.can_edit:
            return

        if value:

            def _start_editing():
                self.set_visible_child_name("editing")
                self._text.grab_focus()

            GLib.timeout_add(START_EDIT_DELAY, _start_editing)
        else:
            self.set_visible_child_name(
                "default" if self.readonly_text else "placeholder"
            )

    _text = Gtk.Template.Child()

    def start_editing(self):
        self.editing = True  # type: ignore[method-assign]

    def done_editing(self, should_commit):
        if self.editing:
            self.emit("done-editing", should_commit)

    @GObject.Signal(name="done-editing", arg_types=(bool,))
    def _done_editing(self, should_commit: bool) -> None:
        self.editing = False  # type: ignore[method-assign]
        if parent := self.get_parent():
            parent.grab_focus()
        if should_commit:
            self.editable_text = self._text.get_buffer().get_text()
        else:
            self._text.get_buffer().set_text(self.editable_text, -1)

    @Gtk.Template.Callback()
    def on_text_focus_out(self, _ctrl):
        if self.editing:
            self.done_editing(True)

    @Gtk.Template.Callback()
    def on_text_key_pressed(self, _ctrl, keyval, _keycode, _state):
        if keyval in (Gdk.KEY_Return, Gdk.KEY_KP_Enter):
            self.done_editing(True)
            return True
        elif keyval == Gdk.KEY_Escape:
            self.done_editing(False)
            return True
        return False


TextField.set_css_name("textfield")



"""Properties used to create the Gaphor data models.

The logic for creating and destroying connections between elements is
implemented in Python property classes. These classes are simply instantiated
like this:
    class Class(Element): pass
    class Comment(Element): pass
    Class.comment = association('comment', Comment,
                                     0, '*', 'annotatedElement')
    Comment.annotatedElement = association('annotatedElement', Element,
                                           0, '*', 'comment')

Same for attributes and enumerations.

Each property type (association, attribute and enumeration) has three specific
methods:
    get():           return the value
    set(value):      set the value or add it to a list
    delete(value=None): delete the value. 'value' is used to tell which value
                      is to be removed (in case of associations with
                      multiplicity > 1).
    load(value):      load 'value' as the current value for this property
    save(save_func):  send the value of the property to save_func(name, value)
"""

from __future__ import annotations

import logging
from typing import (
    Any,
    Callable,
    Generic,
    Iterable,
    Literal,
    Protocol,
    Sequence,
    TypeVar,
    Union,
    overload,
)

from gaphor.core.modeling.collection import collection
from gaphor.core.modeling.event import (
    AssociationAdded,
    AssociationDeleted,
    AssociationSet,
    AssociationUpdated,
    AttributeUpdated,
    DerivedAdded,
    DerivedDeleted,
    DerivedSet,
    DerivedUpdated,
    RedefinedAdded,
    RedefinedDeleted,
    RedefinedSet,
)

__all__ = ["attribute", "enumeration", "association", "derivedunion", "redefine"]


log = logging.getLogger(__name__)


E = TypeVar("E")


class relation_one(Protocol[E]):
    def get(self, obj) -> E | None:
        ...

    def set(self, obj, value: E | None) -> None:
        ...

    def delete(self, obj, value: E | None) -> None:
        ...

    @overload
    def __get__(self, obj: None, class_=None) -> relation_one[E]:
        ...

    @overload
    def __get__(self, obj, class_=None) -> E:
        ...

    def __set__(self, obj, value: E | None) -> None:
        ...

    def __delete__(self, obj) -> None:
        ...


class relation_many(Protocol[E]):
    def get(self, obj) -> collection[E]:
        ...

    def set(self, obj, value: E) -> None:
        ...

    def delete(self, obj, value: E) -> None:
        ...

    @overload
    def __get__(self, obj: None, class_=None) -> relation_many[E]:
        ...

    @overload
    def __get__(self, obj, class_=None) -> collection[E]:
        ...

    def __set__(self, obj, value: E) -> None:
        ...

    def __delete__(self, obj) -> None:
        ...


relation = Union[relation_one, relation_many]

T = TypeVar("T")

Lower = Union[Literal[0], Literal[1], Literal[2]]
Upper = Union[Literal[1], Literal[2], Literal["*"]]


class umlproperty:
    """Superclass for an attribute, enumeration, and association.

    The subclasses should define a ``name`` attribute that contains the name
    of the property. Derived properties (derivedunion and redefine) can be
    connected, they will be notified when the value changes through
    `propagate(self, event)`.

    In some cases properties call out and delegate actions to the ElementFactory,
    for example, in the case of event handling.
    """

    lower: Lower = 0
    upper: Upper = 1

    def __init__(self, name: str):
        self._dependent_properties: set[derived | redefine] = set()
        self.name = name
        self._name = f"_{name}"

    def __get__(self, obj, class_=None):
        return self.get(obj) if obj else self

    def __set__(self, obj, value) -> None:
        self.set(obj, value)

    def __delete__(self, obj) -> None:
        self.delete(obj, None)

    def __repr__(self):
        return str(self)

    def save(self, obj, save_func: Callable[[str, object], None]):
        if hasattr(obj, self._name):
            save_func(self.name, self.get(obj))

    def load(self, obj, value):
        self.set(obj, value)

    def postload(self, obj):
        pass

    def unlink(self, obj):
        """This is called from the Element to denote the element is
        unlinking."""

    def get(self, obj) -> T | None | collection[T]:
        raise NotImplementedError()

    def set(self, obj, value: T | None) -> None:
        raise NotImplementedError()

    def delete(self, obj, value: T | None) -> None:
        raise NotImplementedError()

    def handle(self, event):
        event.element.handle(event)
        for d in self._dependent_properties:
            d.propagate(event)


class attribute(umlproperty, Generic[T]):
    """Attribute.

    Element.attr = attribute('attr', types.StringType, '')
    """

    def __init__(
        self,
        name: str,
        type: type[str | int],
        default: str | int | None = None,
    ):
        super().__init__(name)
        self.type = type
        self.default: str | int | None = default

    def load(self, obj, value: str | None):
        """Load the attribute value."""
        self.set(obj, value)

    def unlink(self, obj):
        self.set(obj, self.default)

    def __str__(self):
        return f"<attribute {self.name}: {self.type} = {self.default}>"

    def get(self, obj):
        return getattr(obj, self._name, self.default)

    def set(self, obj, value):
        if (
            value is not None
            and not isinstance(value, self.type)
            and not isinstance(value, str)
        ):
            raise TypeError(
                f'Value should be of type {hasattr(self.type, "__name__")}'
                and self.type.__name__
                or self.type
            )

        if self.type is int and isinstance(value, (str, bool)):
            value = 0 if value == "False" else 1 if value == "True" else int(value)

        if value == self.get(obj):
            return

        old = self.get(obj)
        if value == self.default and hasattr(obj, self._name):
            delattr(obj, self._name)
        else:
            setattr(obj, self._name, value)
        self.handle(AttributeUpdated(obj, self, old, value))

    def delete(self, obj, value=None):
        old = self.get(obj)
        try:
            delattr(obj, self._name)
        except AttributeError:
            pass
        else:
            self.handle(AttributeUpdated(obj, self, old, self.default))


class enumeration(umlproperty):
    """Enumeration.

      Element.enum = enumeration('enum', ('one', 'two', 'three'), 'one')

    An enumeration is a special kind of attribute that can only hold a
    predefined set of values. Multiplicity is always `[0..1]`
    """

    # All enumerations have a type 'str'
    type = str

    def __init__(self, name: str, values: Sequence[str], default: str):
        super().__init__(name)
        self.values = values
        self.default = default

    def __str__(self):
        return f"<enumeration {self.name}: {self.values} = {self.default}>"

    def get(self, obj):
        return getattr(obj, self._name, self.default)

    def load(self, obj, value: str | None):
        self.set(obj, self.default if value is None else value)

    def unlink(self, obj):
        self.set(obj, self.default)

    def set(self, obj, value):
        if value not in self.values:
            raise TypeError(f"Value should be one of {self.values}")
        old = self.get(obj)
        if value == old:
            return

        if value == self.default:
            delattr(obj, self._name)
        else:
            setattr(obj, self._name, value)
        self.handle(AttributeUpdated(obj, self, old, value))

    def delete(self, obj, value=None):
        old = self.get(obj)
        try:
            delattr(obj, self._name)
        except AttributeError:
            pass
        else:
            self.handle(AttributeUpdated(obj, self, old, self.default))


class association(umlproperty):
    """Association, both uni- and bi-directional.

    Element.assoc = association('assoc', Element, opposite='other')

    We connect a listener to the value added to the association. This will
    cause the association to be ended if we unlink the element on the other end
    of the association.

    If the association is a composite relationship, the association will
    unlink all elements attached to if it is unlinked.
    """

    def __init__(
        self,
        name: str,
        type: type,
        lower: Lower = 0,
        upper: Upper = "*",
        composite: bool = False,
        opposite: str | None = None,
    ):
        super().__init__(name)
        self.type = type
        self.lower = lower
        self.upper = upper
        self.composite = composite
        self.opposite = opposite
        self.stub: associationstub | None = None

    def save(self, obj, save_func: Callable[[str, object], None]):
        if hasattr(obj, self._name):
            if v := self.get(obj):
                save_func(self.name, v)

    def load(self, obj, value):
        if self.opposite:
            # Loading should not steal references from other elements
            opposite = getattr(type(value), self.opposite)
            if opposite.upper == 1 and (opval := opposite.get(value)):
                if opval is not obj:
                    log.debug(f"Cannot steal reference from {value}")
                return

        self.set(obj, value)

    def __str__(self):
        if self.lower == self.upper:
            s = f"<association {self.name}: {self.type.__name__}[{self.lower}]"
        else:
            s = f"<association {self.name}: {self.type.__name__}[{self.lower}..{self.upper}]"
        if self.opposite:
            s += f" {self.composite and '<>' or ''}-> {self.opposite}"
        return f"{s}>"

    def get(self, obj):
        return self._get_one(obj) if self.upper == 1 else self._get_many(obj)

    def _get_one(self, obj) -> T | None:
        return getattr(obj, self._name, None)

    def _get_many(self, obj) -> collection[T]:
        v: collection[T] | None = getattr(obj, self._name, None)
        if v is None:
            # Create the empty collection here since it may
            # be used to add.
            v = collection(self, obj, self.type)
            setattr(obj, self._name, v)
        return v

    def set(
        self, obj, value: T | None, index: int | None = None, from_opposite=False
    ) -> None:
        """Set a new value for our attribute. If this is a collection, append
        to the existing collection.

        This method is called from the opposite association property.
        """
        if obj is value:
            raise TypeError(f"Can not set {obj}.{self.name} to itself")

        if self.upper == 1:
            self._set_one(obj, value, from_opposite)
        else:
            self._set_many(obj, value, index, from_opposite)

    def _set_one(self, obj, value, from_opposite=False) -> None:
        if not (isinstance(value, self.type) or (value is None)):
            raise TypeError(
                f"Value should be of type {self.type.__name__}, got a {type(value)} instead"
            )

        old = self._get_one(obj)

        # do nothing if we are assigned our current value:
        # Still do your thing, since undo handlers expect that.
        if value is old:
            return

        if old:
            self._del_one(obj, old, do_notify=False)

        if value is not None:
            setattr(obj, self._name, value)
            try:
                self._set_opposite(obj, value, from_opposite)
            except Exception:
                setattr(obj, self._name, old)
                raise

        self.handle(AssociationSet(obj, self, old, value))

    def _set_many(
        self, obj, value, index, from_opposite=False, from_load=False
    ) -> None:
        if not isinstance(value, self.type):
            raise TypeError(f"Value should be of type {self.type.__name__}")

        # Set the actual value
        c: collection = self._get_many(obj)
        if value in c:
            if from_load:
                c.items.remove(value)
                if index is None:
                    c.items.append(value)
                else:
                    c.items.insert(index, value)
            return

        if index is None:
            c.items.append(value)
        else:
            c.items.insert(index, value)

        try:
            self._set_opposite(obj, value, from_opposite)
        except Exception:
            if value in c.items:
                c.items.remove(value)
            raise

        self.handle(AssociationAdded(obj, self, value))

    def _set_opposite(self, obj, value: T | None, from_opposite=False) -> None:
        if not from_opposite and self.opposite:
            opposite = getattr(type(value), self.opposite)
            if not opposite.opposite:
                opposite.stub = self
            opposite.set(value, obj, from_opposite=True)
        elif not self.opposite:
            if not self.stub:
                self.stub = associationstub(self)
                # Do not let property start with underscore, or it will not be found
                # as an umlproperty.
                setattr(self.type, "GAPHOR__associationstub__%x" % id(self), self.stub)
            self.stub.set(value, obj)

    def delete(self, obj, value, from_opposite=False, do_notify=True):
        """Delete is used for element deletion and for removal of elements from
        a list."""
        if self.upper == 1:
            self._del_one(obj, self._get_one(obj), from_opposite, do_notify)
        else:
            self._del_many(obj, value, from_opposite, do_notify)

    def _del_one(self, obj, value, from_opposite=False, do_notify=True):
        if value is None:
            return

        self._del_opposite(obj, value, from_opposite)

        try:
            delattr(obj, self._name)
        except AttributeError:
            pass
        else:
            if do_notify:
                self.handle(AssociationSet(obj, self, value, None))

    def _del_many(self, obj, value, from_opposite=False, do_notify=True):
        if not value:
            raise Exception("Can not delete collections")

        self._del_opposite(obj, value, from_opposite)

        c: collection
        if c := self._get_many(obj):
            items: list = c.items
            try:
                index = items.index(value)
                items.remove(value)
            except ValueError:
                pass
            else:
                if do_notify:
                    self.handle(AssociationDeleted(obj, self, value, index))

            # Remove items collection if empty
            if not items:
                delattr(obj, self._name)

    def _del_opposite(self, obj, value, from_opposite):
        if not from_opposite and self.opposite:
            getattr(type(value), self.opposite).delete(value, obj, from_opposite=True)
        elif not self.opposite:
            if self.stub:
                self.stub.delete(value, obj, from_opposite=True)

    def unlink(self, obj):
        if values := self.get(obj):
            if self.upper == 1:
                values = [values]

            composite = self.composite
            for value in list(values):
                self.delete(obj, value)
                if composite:
                    value.unlink()


class AssociationStubError(Exception):
    pass


class associationstub(umlproperty):
    """An association stub is an internal thingy that ensures all associations
    are always bidirectional.

    This helps the application when one end of the association is
    unlinked. On unlink() of an element all `umlproperty`'s are iterated
    and called by their unlink() method.
    """

    def __init__(self, association: association):
        super().__init__("stub_%x" % id(self))
        self.association = association

    def __get__(self, obj, class_=None):
        if obj:
            raise AssociationStubError("getting values not allowed")
        return self

    def __set__(self, obj, value):
        raise AssociationStubError("setting values not allowed")

    def __delete__(self, obj):
        raise AssociationStubError("deleting values not allowed")

    def save(self, obj, save_func):
        pass

    def load(self, obj, value):
        pass

    def unlink(self, obj):
        values = getattr(obj, self._name, [])
        for value in set(values):
            self.association.delete(value, obj)

    def set(self, obj, value):
        try:
            getattr(obj, self._name).add(value)
        except AttributeError:
            setattr(obj, self._name, {value})

    def delete(self, obj, value, from_opposite=False):
        try:
            c = getattr(obj, self._name)
        except AttributeError:
            pass
        else:
            c.discard(value)


class unioncache:
    """Small cache helper object for derivedunions."""

    def __init__(self, owner: object, data: object, version: int) -> None:
        self.owner = owner
        self.data = data
        self.version = version


class derived(umlproperty, Generic[T]):
    """Base class for derived properties, both derived unions and custom
    properties.

    Note that, although this derived property sends DerivedAdded,
    -Delete- and Set events, this gives just an assumption that something
    may have changed. If something actually changed depends on the filter
    applied to the derived property.

    NB. filter returns a *list* of filtered items, even when upper bound is 1.
    """

    opposite = None

    def __init__(  # type: ignore[misc]
        self,
        name: str,
        type: type[T],
        lower: Lower,
        upper: Upper,
        filter: Callable[[Any], list[T | None]],
        *subsets: relation,
    ) -> None:
        super().__init__(name)
        self.version = 1
        self.type = type
        self.lower = lower
        self.upper = upper
        self.filter = filter
        self.subsets: set[umlproperty] = set()

        for s in subsets:
            self.add(s)

    def add(self, subset):
        self.subsets.add(subset)
        assert isinstance(
            subset, (association, derived)
        ), f"have element {subset}, expected association"
        subset._dependent_properties.add(self)

    def load(self, obj, value):
        raise ValueError(
            f"Derivedunion: Properties should not be loaded in a derived union {self.name}: {value}"
        )

    def postload(self, obj):
        self.version += 1

    def save(self, obj, save_func):
        pass

    def __str__(self):
        return f"<derived {self.name}: {self.type} {str(list(map(str, self.subsets)))[1:-1]}>"

    def _update(self, obj):
        """Update the list of items.

        Returns an unioncache instance.
        """
        u = self.filter(obj)
        if self.upper == 1:
            uc = unioncache(self, u[0] if u else None, self.version)
        else:
            c = collection(self, obj, self.type)
            c.items.extend(u)  # type: ignore[arg-type]
            uc = unioncache(self, c, self.version)
        setattr(obj, self._name, uc)
        return uc

    def get(self, obj):
        if self.subsets:
            try:
                uc = getattr(obj, self._name)
                if uc.version != self.version:
                    uc = self._update(obj)
                assert self is uc.owner
            except AttributeError:
                uc = self._update(obj)
        else:
            uc = self._update(obj)
        return uc.data

    def set(self, obj, value):
        raise AttributeError(f"Can not set values on union {self.name}: {self.type}")

    def delete(self, obj, value=None):
        raise AttributeError(f"Can not delete values on union {self.name}: {self.type}")

    def propagate(self, event):
        """Re-emit state change for the derived properties as Derived*Event's.

        If multiplicity is [0..1]:   send DerivedSet if len(union) < 2
        if multiplicity is [*]:   send DerivedAdded and DerivedDeleted
        if value not in derived union and
        """
        if event.property not in self.subsets:
            return
        if not isinstance(event, AssociationUpdated):
            return

        # mimic the events for Set/Add/Delete
        if self.upper == 1:
            old_value = hasattr(event.element, self._name) and self.get(event.element)
            # Make sure unions are created again
            self.version += 1
            new_value = self.get(event.element)
            if old_value != new_value:
                self.handle(DerivedSet(event.element, self, old_value, new_value))
        else:
            # Make sure unions are created again
            self.version += 1

            if isinstance(event, AssociationSet):
                self.handle(DerivedDeleted(event.element, self, event.old_value))
                self.handle(DerivedAdded(event.element, self, event.new_value))

            elif isinstance(event, AssociationAdded):
                self.handle(DerivedAdded(event.element, self, event.new_value))

            elif isinstance(event, AssociationDeleted):
                self.handle(DerivedDeleted(event.element, self, event.old_value))

            elif isinstance(event, AssociationUpdated):
                self.handle(DerivedUpdated(event.element, self))
            else:
                log.error(
                    "Don't know how to handle event "
                    + str(event)
                    + " for derived union"
                )


def object_has_property(obj, prop):
    found = getattr(type(obj), prop.name, None)
    while isinstance(found, redefine):
        found = found.original
    return prop is found


class derivedunion(derived[T]):
    """Derived union.

      Element.union = derivedunion('union', subset1, subset2..subsetn)

    The subsets are the properties that participate in the union (Element.name).
    """

    def __init__(
        self,
        name: str,
        type: type[T],
        lower: Lower = 0,
        upper: Upper = "*",
        *subsets: relation,
    ):
        super().__init__(name, type, lower, upper, self._union, *subsets)

    def _union(self, obj, exclude=None):
        """Returns a union of all values as a set."""
        u: set[T] = set()
        for s in self.subsets:
            if s is exclude or not object_has_property(obj, s):
                continue

            tmp: Iterable[T] | T | None = s.get(obj)
            if isinstance(tmp, Iterable):
                u.update(tmp)
            elif tmp:
                # [0..1] property
                u.add(tmp)
        return list(u)

    def propagate(self, event):
        """Re-emit state change for the derived union (as Derived*Event's).

        If multiplicity is [0..1]:   send DerivedSet if len(union) < 2
        if multiplicity is [*]:   send DerivedAdded and DerivedDeleted
        if value not in derived union and
        """
        if event.property not in self.subsets:
            return
        # Make sure unions are created again
        self.version += 1

        if not isinstance(event, AssociationUpdated):
            return

        values = set(self._union(event.element, exclude=event.property))

        if self.upper == 1:
            assert isinstance(event, AssociationSet)
            old_value, new_value = event.old_value, event.new_value
            # This is a [0..1] event
            if len(self.subsets) > 1:
                new_values = set(values)
                if new_value:
                    new_values.add(new_value)
                if len(new_values) > 1:
                    # In an in-between state. Do not emit notifications
                    return
                if values:
                    new_value = next(iter(values))
            # Only one subset element, so pass the values on
            self.handle(DerivedSet(event.element, self, old_value, new_value))
        elif isinstance(event, AssociationSet):
            old_value, new_value = event.old_value, event.new_value
            if old_value and old_value not in values:
                self.handle(DerivedDeleted(event.element, self, old_value))
            if new_value and new_value not in values:
                self.handle(DerivedAdded(event.element, self, new_value))

        elif isinstance(event, AssociationAdded):
            new_value = event.new_value
            if new_value not in values:
                self.handle(DerivedAdded(event.element, self, new_value))

        elif isinstance(event, AssociationDeleted):
            old_value = event.old_value
            if old_value not in values:
                self.handle(DerivedDeleted(event.element, self, old_value))

        elif isinstance(event, AssociationUpdated):
            self.handle(DerivedUpdated(event.element, self))
        else:
            log.error(f"Don't know how to handle event {event} for derived union")


class redefine(umlproperty):
    """Redefined association.

      Element.redefine = redefine(Element, 'redefine', Class, Element.assoc)

    If the redefine eclipses the original property (it has the same name)
    it ensures that the original values are saved and restored.
    """

    def __init__(
        self,
        decl_class: type[E],
        name: str,
        type: type[T],
        original: relation,
    ):
        super().__init__(name)
        assert isinstance(
            original, (association, derived)
        ), f"expected association or derived, got {original}"
        self.decl_class = decl_class
        self.type = type
        self.original: association | derived = original
        self.upper = original.upper
        self.lower = original.lower

        original._dependent_properties.add(self)

    @property
    def opposite(self) -> str | None:
        return (
            self.original.opposite if isinstance(self.original, association) else None
        )

    @property
    def composite(self) -> bool:
        return (
            self.original.composite if isinstance(self.original, association) else False
        )

    def load(self, obj, value):
        if self.original.name == self.name:
            self.original.load(obj, value)

    def postload(self, obj):
        if self.original.name == self.name:
            self.original.postload(obj)

    def save(self, obj, save_func):
        if self.original.name == self.name:
            self.original.save(obj, save_func)

    def unlink(self, obj):
        if self.original.name == self.name:
            self.original.unlink(obj)

    def __str__(self) -> str:
        return f"<redefine {self.name}[{self.lower}..{self.upper}]: {self.type.__name__} = {self.original}>"

    def get(self, obj):
        return self.original.get(obj)

    def set(self, obj, value, from_opposite=False):
        if not (isinstance(value, self.type) or (self.upper == 1 and value is None)):
            raise TypeError(
                f"Value should be of type {self.type.__name__}, got a {type(value)} instead"
            )
        assert isinstance(self.original, association)
        return self.original.set(obj, value, from_opposite=from_opposite)

    def delete(self, obj, value, from_opposite=False, do_notify=True):
        assert isinstance(self.original, association)
        return self.original.delete(obj, value, from_opposite, do_notify)

    def propagate(self, event):
        if event.property is self.original and isinstance(
            event.element, self.decl_class
        ):
            # mimic the events for Set/Add/Delete
            if isinstance(event, AssociationSet):
                self.handle(
                    RedefinedSet(event.element, self, event.old_value, event.new_value)
                )
            elif isinstance(event, AssociationAdded):
                self.handle(RedefinedAdded(event.element, self, event.new_value))
            elif isinstance(event, AssociationDeleted):
                self.handle(
                    RedefinedDeleted(event.element, self, event.old_value, event.index)
                )
            else:
                log.error(
                    "Don't know how to handle event "
                    + str(event)
                    + " for redefined association"
                )



"""The element editor is a utility window used for editing elements."""

import importlib.resources
import logging
from typing import Optional
from unicodedata import normalize

from babel import Locale
from gi.repository import Adw, GLib, Gtk, GtkSource

from gaphor.abc import ActionProvider
from gaphor.core import Transaction, action, event_handler
from gaphor.core.modeling import Presentation, StyleSheet
from gaphor.core.modeling.event import (
    AssociationUpdated,
    AttributeUpdated,
    ElementCreated,
    ModelReady,
)
from gaphor.diagram.propertypages import PropertyPages, new_resource_builder
from gaphor.event import ModelLoaded
from gaphor.i18n import localedir
from gaphor.ui.abc import UIComponent
from gaphor.ui.csscompletion import (
    CompletionProviderWrapper,
    CssFunctionProposals,
    CssNamedColorProposals,
    CssPropertyProposals,
)
from gaphor.ui.event import DiagramSelectionChanged
from gaphor.ui.modelmerge import ModelMerge

log = logging.getLogger(__name__)

new_builder = new_resource_builder("gaphor.ui", "elementeditor")


class DelayedFunction:
    def __init__(self, delay, function):
        self._delay = delay
        self._function = function
        self._source_id = 0

    def __call__(self, *args):
        if self._source_id:
            GLib.source_remove(self._source_id)

        def timeout_function():
            try:
                self._function(*args)
            finally:
                self._source_id = 0

        self._source_id = GLib.timeout_add(self._delay, timeout_function)


class ElementEditor(UIComponent, ActionProvider):
    """The ElementEditor class is a utility window used to edit UML elements.

    It will display the properties of the currently selected element in
    the diagram.
    """

    def __init__(
        self, event_manager, element_factory, modeling_language, diagrams, properties
    ):
        """Constructor.

        Build the action group for the element editor window. This will
        place a button for opening the window in the toolbar. The widget
        attribute is a PropertyEditor.
        """
        self.event_manager = event_manager
        self.properties = properties
        self.editors = EditorStack(event_manager, diagrams, properties)
        self.preferences = PreferencesStack(event_manager, element_factory)
        self.modelmerge = ModelMerge(event_manager, element_factory, modeling_language)

    def open(self):
        """Display the ElementEditor pane."""
        builder = new_builder("elementeditor", "style-sheet-buffer")

        self.revealer = builder.get_object("elementeditor")
        self.revealer.set_reveal_child(self.properties.get("show-editors", True))
        self.editor_stack = builder.get_object("editor-stack")

        resolve = builder.get_object("modelmerge-resolve")
        resolve.connect_after("clicked", self.on_model_loaded)

        self.editors.open(builder)
        self.preferences.open(builder)
        self.modelmerge.open(builder)

        self.event_manager.subscribe(self.on_model_loaded)

        self.on_model_loaded(None)

        return self.revealer

    def close(self, widget=None):
        """Hide the element editor window and deactivate the toolbar button.

        Both the widget and event parameters default to None and are
        idempotent if set.
        """

        self.event_manager.unsubscribe(self.on_model_loaded)

        self.editors.close()
        self.preferences.close()
        self.modelmerge.close()

        return True

    @event_handler(ModelLoaded)
    def on_model_loaded(self, event):
        self.editor_stack.set_visible_child_name(
            "modelmerge" if self.modelmerge.needs_merge else "editors"
        )

    @action(
        name="show-editors",
        shortcut="F9",
        state=lambda self: self.properties.get("show-editors", True),
    )
    def toggle_editor_visibility(self, active):
        self.revealer.set_reveal_child(active)
        self.properties.set("show-editors", active)

    @action(name="show-preferences", shortcut="<Primary>comma", state=False)
    def toggle_editor_preferences(self, active):
        if not self.revealer.get_child_revealed():
            self.revealer.activate_action("win.show-editors", None)

        self.editor_stack.set_visible_child_name(
            "preferences"
            if active
            else "modelmerge"
            if self.modelmerge.needs_merge
            else "editors"
        )

    @action(
        name="reset-tool-after-create",
        state=lambda self: self.properties.get("reset-tool-after-create", True),
    )
    def reset_tool_after_create(self, active):
        self.properties.set("reset-tool-after-create", active)

    @action(
        name="remove-unused-elements",
        state=lambda self: self.properties.get("remove-unused-elements", True),
    )
    def remove_unused_elements(self, active):
        self.properties.set("remove-unused-elements", active)


class EditorStack:
    def __init__(self, event_manager, diagrams, properties):
        self.event_manager = event_manager
        self.diagrams = diagrams
        self.properties = properties

        self.vbox: Optional[Gtk.Box] = None
        self._current_item = None
        self._expanded_pages = {}

    def open(self, builder):
        """Display the ElementEditor pane."""
        self.vbox = builder.get_object("editors")

        current_view = self.diagrams.get_current_view()
        self._selection_changed(
            focused_item=current_view and current_view.selection.focused_item
        )

        self.event_manager.subscribe(self._selection_changed)
        self.event_manager.subscribe(self._element_changed)

    def close(self):
        self.event_manager.unsubscribe(self._selection_changed)
        self.event_manager.unsubscribe(self._element_changed)

        self.vbox = None
        self._current_item = None

    def _get_adapters(self, item):
        """Return an ordered list of (order, name, adapter)."""
        adaptermap = {}
        if isinstance(item, Presentation) and item.subject:
            for adapter in PropertyPages(item.subject):
                adaptermap[(adapter.order, adapter.__class__.__name__)] = adapter
        for adapter in PropertyPages(item):
            adaptermap[(adapter.order, adapter.__class__.__name__)] = adapter

        return sorted(adaptermap.items())

    def create_pages(self, item):
        """Load all tabs that can operate on the given item."""
        assert self.vbox
        adapters = self._get_adapters(item)

        for (_, name), adapter in adapters:
            try:
                page = adapter.construct()
                if not page:
                    continue
                elif isinstance(page, Gtk.Expander):
                    page.set_expanded(self._expanded_pages.get(name, True))
                    page.connect_after("activate", self.on_expand, name)
                self.vbox.append(page)
            except Exception:
                log.error(
                    f"Could not construct property page for {name}", exc_info=True
                )

    def clear_pages(self):
        """Remove all tabs from the notebook."""
        assert self.vbox
        while page := self.vbox.get_first_child():
            self.vbox.remove(page)

    def on_expand(self, widget, name):
        self._expanded_pages[name] = widget.get_expanded()

    @event_handler(DiagramSelectionChanged)
    def _selection_changed(self, event=None, focused_item=None):
        """Called when a diagram item receives focus.

        This reloads all tabs based on the current selection.
        """
        assert self.vbox
        item = event and event.focused_item or focused_item
        if item is self._current_item and self.vbox.get_first_child():
            return

        self._current_item = item
        self.clear_pages()

        if item:
            self.create_pages(item)
        else:
            self.show_no_item_selected()

    def show_no_item_selected(self):
        assert self.vbox
        builder = new_builder("no-item-selected")

        self.vbox.append(builder.get_object("no-item-selected"))

        tips = builder.get_object("tips")

        def on_show_tips_changed(checkbox, gparam):
            active = checkbox.get_active()
            tips.set_visible(active)
            self.properties.set("show-tips", active)

        show_tips = builder.get_object("show-tips")
        show_tips.connect("notify::active", on_show_tips_changed)
        show_tips.set_active(self.properties.get("show-tips", True))
        on_show_tips_changed(show_tips, None)

    @event_handler(AssociationUpdated)
    def _element_changed(self, event: AssociationUpdated):
        if event.property is Presentation.subject:  # type: ignore[misc] # noqa: F821
            element = event.element
            if element is self._current_item:
                self.clear_pages()
                self.create_pages(self._current_item)


class PreferencesStack:
    """Support code for the Preferences (cog) pane."""

    def __init__(self, event_manager, element_factory):
        self.event_manager = event_manager
        self.element_factory = element_factory
        self.lang_manager = GtkSource.LanguageManager.get_default()
        self.style_manager = Adw.StyleManager.get_default()
        self._notify_dark_id = 0
        self.lang_manager.append_search_path(
            str(importlib.resources.files("gaphor") / "ui" / "language-specs")
        )

        def tx_update_style_sheet(style_sheet, text):
            self._in_update = 1
            with Transaction(event_manager):
                style_sheet.styleSheet = text
            self._in_update = 0

        self._style_sheet_update = DelayedFunction(800, tx_update_style_sheet)
        self._in_update = 0

        self._languages = [
            (None, "English"),
            *sorted(
                (
                    (d.name, Locale.parse(d.name).get_language_name().title())
                    for d in localedir.iterdir()
                    if d.is_dir()
                ),
                key=lambda k: normalize("NFC", k[1]).casefold(),
            ),
        ]

    def open(self, builder):
        self.language_dropdown = builder.get_object("language-dropdown")
        self.style_sheet_buffer = builder.get_object("style-sheet-buffer")
        self.style_sheet_view = builder.get_object("style-sheet-view")

        self.style_sheet_buffer.set_language(
            self.lang_manager.get_language("gaphorcss")
        )

        view_completion = self.style_sheet_view.get_completion()
        view_completion.add_provider(
            CompletionProviderWrapper(priority=3, proposals=CssFunctionProposals())
        )
        view_completion.add_provider(
            CompletionProviderWrapper(priority=-4, proposals=CssNamedColorProposals())
        )
        view_completion.add_provider(
            CompletionProviderWrapper(priority=1, proposals=CssPropertyProposals())
        )
        assert self.style_manager
        self._notify_dark_id = self.style_manager.connect(
            "notify::dark", self._on_notify_dark
        )
        self._on_notify_dark(self.style_manager, None)

        language_model = builder.get_object("language-model")
        for _code, language in self._languages:
            language_model.append(language)

        self.event_manager.subscribe(self._model_ready)
        self.event_manager.subscribe(self._style_sheet_created)
        self.event_manager.subscribe(self._style_sheet_changed)
        self.style_sheet_buffer.connect("changed", self.on_style_sheet_changed)
        self.language_dropdown.connect("notify::selected", self.on_language_changed)

        self.update()

    def close(self):
        self.event_manager.unsubscribe(self._model_ready)
        self.event_manager.unsubscribe(self._style_sheet_changed)
        self.event_manager.unsubscribe(self._style_sheet_created)
        if self._notify_dark_id and self.style_manager:
            self._notify_dark_id = self.style_manager.disconnect(self._notify_dark_id)

    @property
    def style_sheet(self):
        return next(self.element_factory.select(StyleSheet), None)

    def on_style_sheet_changed(self, buffer):
        style_sheet = self.style_sheet
        if style_sheet and not self._in_update:
            text = buffer.get_text(
                buffer.get_start_iter(), buffer.get_end_iter(), False
            )

            self._style_sheet_update(style_sheet, text)

    def update(self):
        style_sheet = self.style_sheet
        if not style_sheet or self._in_update:
            return

        self._in_update = 1
        self.style_sheet_buffer.set_text(style_sheet.styleSheet or "")
        self.language_dropdown.set_selected(
            next(
                (
                    i
                    for i, (code, _) in enumerate(self._languages)
                    if code == style_sheet.naturalLanguage
                ),
                0,
            )
        )
        self._in_update = 0

    @event_handler(ModelReady)
    def _model_ready(self, event):
        self.update()

    @event_handler(ElementCreated)
    def _style_sheet_created(self, event):
        if isinstance(event.element, StyleSheet):
            self.update()

    @event_handler(AttributeUpdated)
    def _style_sheet_changed(self, event):
        if event.property is StyleSheet.styleSheet:
            self.update()

    def on_language_changed(self, dropdown, _gparam):
        if (style_sheet := self.style_sheet) and not self._in_update:
            with Transaction(self.event_manager):
                code, _name = self._languages[dropdown.get_selected()]
                style_sheet.naturalLanguage = code

    def _on_notify_dark(self, style_manager, _gparam):
        scheme_manager = GtkSource.StyleSchemeManager.get_default()
        scheme = scheme_manager.get_scheme(
            "Adwaita-dark" if style_manager.get_dark() else "Adwaita"
        )
        self.style_sheet_buffer.set_style_scheme(scheme)



"""The core modeling events."""
from gaphor.event import ServiceEvent


class RevertibleEvent:
    """Base type for all events that can be reversed.

    This event can be used as "low level" event for anything that should
    be revertible/undoable.
    """

    requires_transaction = True

    def __init__(self, element):
        self.element = element

    def revert(self, target):
        """Reverse whatever caused the event.

        `target` Is the element the action should be performed upon,
        which may be a different element than the one that caused the
        event.
        """
        raise NotImplementedError("Method {self}.revert() has not been implemented")


class ElementUpdated:
    """Generic event fired when element state changes."""

    def __init__(self, element, property):
        self.element = element
        self.property = property


class AttributeUpdated(ElementUpdated):
    """An attribute has changed value."""

    def __init__(self, element, attribute, old_value, new_value):
        """Constructor.

        The element parameter is the element with the changing
        attribute.  The attribute parameter is the parameter element
        that changed.  The old_value is the old value of the attribute
        and the new_value is the new value of the attribute.
        """

        super().__init__(element, attribute)
        self.old_value = old_value
        self.new_value = new_value


class AssociationUpdated(ElementUpdated):
    """An association element has changed."""

    def __init__(self, element, association):
        """Constructor.

        The element parameter is the element the association is changing
        from.  The association parameter is the changed association
        element.
        """

        super().__init__(element, association)


class AssociationSet(AssociationUpdated):
    """An association element has been set."""

    def __init__(self, element, association, old_value, new_value):
        """Constructor.

        The element parameter is the element setting the association
        element.  The association parameter is the association element
        being set.  The old_value parameter is the old association and
        the new_value parameter is the new association.
        """

        super().__init__(element, association)
        self.old_value = old_value
        self.new_value = new_value


class AssociationAdded(AssociationUpdated):
    """An association element has been added."""

    def __init__(self, element, association, new_value):
        """Constructor.

        The element parameter is the element the association has been
        added to.  The association parameter is the association element
        being added.
        """

        super().__init__(element, association)
        self.new_value = new_value


class AssociationDeleted(AssociationUpdated):
    """An association element has been deleted."""

    def __init__(self, element, association, old_value, index):
        """Constructor.

        The element parameter is the element the association has been
        deleted from.  The association parameter is the deleted
        association element.
        """

        super().__init__(element, association)
        self.old_value = old_value
        self.index = index


class DerivedUpdated(AssociationUpdated):
    """A derived property has changed."""


class DerivedSet(AssociationSet, DerivedUpdated):
    """A generic derived set event."""

    def __init__(self, element, association, old_value, new_value):
        """Constructor.

        The element parameter is the element to which the derived set
        belongs.  The association parameter is the association of the
        derived set.
        """

        super().__init__(element, association, old_value, new_value)


class DerivedAdded(AssociationAdded, DerivedUpdated):
    """A derived property has been added."""

    def __init__(self, element, association, new_value):
        """Constructor.

        The element parameter is the element to which the derived
        property belongs.  The association parameter is the association
        of the derived property.
        """

        super().__init__(element, association, new_value)


class DerivedDeleted(AssociationDeleted, DerivedUpdated):
    """A derived property has been deleted."""

    def __init__(self, element, association, old_value):
        """Constructor.

        The element parameter is the element to which the derived
        property belongs.  The association parameter is the association
        of the derived property.
        """

        super().__init__(element, association, old_value, index=-1)


class RedefinedSet(AssociationSet):
    """A redefined property has been set."""

    def __init__(self, element, association, old_value, new_value):
        """Constructor.

        The element parameter is the element to which the property
        belongs.  The association parameter is association of the
        property.
        """

        super().__init__(element, association, old_value, new_value)


class RedefinedAdded(AssociationAdded):
    """A redefined property has been added."""

    def __init__(self, element, association, new_value):
        """Constructor.

        The element parameter is the element to which the property
        belongs.  The association parameter is the association of the
        property.
        """

        super().__init__(element, association, new_value)


class RedefinedDeleted(AssociationDeleted):
    """A redefined property has been deleted."""

    def __init__(self, element, association, old_value, index):
        """Constructor.

        The element parameter is the element to which the property
        belongs.  The association parameter is the association of the
        property.
        """

        super().__init__(element, association, old_value, index)


class ElementCreated(ServiceEvent):
    """An element has been created."""

    def __init__(self, service, element, diagram=None):
        """Constructor.

        The service parameter is the service responsible for creating
        the element.  The element parameter is the element being
        created.
        """
        super().__init__(service)
        self.element = element
        self.diagram = diagram


class ElementDeleted(ServiceEvent):
    """An element has been deleted."""

    def __init__(self, service, element, diagram=None):
        """Constructor.

        The service parameter is the service responsible for deleting
        the element.  The element parameter is the element being
        deleted.
        """
        super().__init__(service)
        self.element = element
        self.diagram = diagram


class ModelReady(ServiceEvent):
    """A generic element factory event."""

    def __init__(self, service):
        """Constructor.

        The service parameter is the service the emitted the event.
        """
        super().__init__(service)


class ModelFlushed(ServiceEvent):
    """The element factory has been flushed."""

    def __init__(self, service):
        """Constructor.

        The service parameter is the service responsible for flushing
        the factory.
        """
        super().__init__(service)



"""With `icon_name` you can retrieve an icon name for a model element."""

import re
from functools import singledispatch

TO_KEBAB = re.compile(r"([a-z0-9])([A-Z]+)")


def to_kebab_case(s):
    return TO_KEBAB.sub("\\1-\\2", s).lower()


def get_default_icon_name(element):
    """Get an icon name for a model element."""
    return f"gaphor-{to_kebab_case(element.__class__.__name__)}-symbolic"


icon_name = singledispatch(get_default_icon_name)



"""Support for actions in generic files."""

from __future__ import annotations

from typing import Any, Callable, Iterator, get_type_hints


class action:
    """Decorator. Turns a regular function (/method) into a full blown Action
    class.

    >>> class A:
    ...     @action(name="my_action", label="my action")
    ...     def myaction(self):
    ...         print("action called")
    >>> a = A()
    >>> a.myaction()
    action called
    >>> is_action(a.myaction)
    True
    >>> for method in dir(A):
    ...     if is_action(getattr(A, method, None)):
    ...         print(method)
    myaction
    >>> A.myaction.__action__.name
    'my_action'
    >>> A.myaction.__action__.label
    'my action'
    """

    def __init__(  # type: ignore[misc]
        self,
        name,
        label: str | None = None,
        tooltip: str | None = None,
        icon_name: str | None = None,
        shortcut: str | tuple[str, ...] | None = None,
        state: bool | Callable[[Any], bool | str] | None = None,
    ):
        self.scope, self.name = name.split(".", 2) if "." in name else ("win", name)
        self.label = label
        self.tooltip = tooltip
        self.icon_name = icon_name
        self.shortcut = shortcut
        self.state = state
        self.arg_type = None

    @property
    def detailed_name(self) -> str:
        return f"{self.scope}.{self.name}"

    @property
    def shortcuts(self) -> Iterator[str]:
        s = self.shortcut
        if isinstance(s, str):
            yield s
        elif s:
            yield from s

    def __call__(self, func):
        type_hints = get_type_hints(func)
        if "return" in type_hints:
            del type_hints["return"]
        if type_hints:
            # assume the first argument (excluding self) is our parameter
            self.arg_type = next(iter(type_hints.values()))
        func.__action__ = self
        return func


def is_action(func):
    return hasattr(func, "__action__")



from gaphas.view import GtkView
from gi.repository import Gdk, Gtk

from gaphor.core import Transaction
from gaphor.core.modeling import Presentation, self_and_owners


def shortcut_tool(event_manager):
    ctrl = Gtk.EventControllerKey.new()
    ctrl.connect("key-pressed", on_delete, event_manager)
    return ctrl


def on_delete(ctrl, keyval, keycode, state, event_manager):
    """Handle the 'Delete' key.

    This can not be handled directly (through GTK's accelerators),
    otherwise this key will confuse the text edit stuff.
    """
    view: GtkView = ctrl.get_widget()
    if keyval in (Gdk.KEY_Delete, Gdk.KEY_BackSpace):
        delete_selected_items(view, event_manager)
        return True
    return False


def delete_selected_items(view: GtkView, event_manager):
    with Transaction(event_manager):
        items = view.selection.selected_items
        for i in list(items):
            assert isinstance(i, Presentation)
            if i.subject and i.subject in self_and_owners(i.diagram):
                del i.diagram.element
            i.unlink()



"""State transition implementation."""

from gaphor import UML
from gaphor.diagram.presentation import LinePresentation, Named
from gaphor.diagram.shapes import Box, Text, draw_arrow_tail
from gaphor.diagram.support import represents
from gaphor.UML.recipes import stereotypes_str


@represents(UML.Transition, head=UML.Transition.source, tail=UML.Transition.target)
class TransitionItem(Named, LinePresentation[UML.Transition]):
    """Representation of state transition."""

    def __init__(self, diagram, id=None):
        super().__init__(
            diagram,
            id,
            shape_middle=Text(
                text=lambda: self.subject
                and self.subject.guard
                and self.subject.guard.specification
                and f"[{self.subject.guard.specification}]"
                or ""
            ),
            shape_tail=Box(
                Text(
                    text=lambda: stereotypes_str(self.subject),
                ),
                Text(text=lambda: self.subject.name or ""),
            ),
        )

        self.watch("subject[NamedElement].name")
        self.watch("subject.appliedStereotype.classifier.name")

        self.watch("subject[Transition].guard[Constraint].specification")

        self.draw_tail = draw_arrow_tail



from functools import singledispatch

from gaphor.core.modeling.coremodel import (
    ElementChange,
    RefChange,
    ValueChange,
)


@singledispatch
def applicable(change, element_factory) -> bool:
    raise NotImplementedError


@singledispatch
def apply_change(change, element_factory, modeling_factory):
    raise NotImplementedError


@applicable.register
def _(change: ElementChange, element_factory) -> bool:
    if change.diagram_id and not element_factory.lookup(change.diagram_id):
        return False
    element = element_factory.lookup(change.element_id)
    return not bool(element) if change.op == "add" else bool(element)


@apply_change.register
def _(change: ElementChange, element_factory, modeling_factory):
    if change.applied:
        return
    if change.op == "add":
        element_type = modeling_factory.lookup_element(change.element_name)
        diagram = (
            element_factory.lookup(change.diagram_id) if change.diagram_id else None
        )
        element_factory.create_as(element_type, change.element_id, diagram=diagram)
    elif change.op == "remove":
        if element := element_factory.lookup(change.element_id):
            element.unlink()
    change.applied = True


@applicable.register
def _(change: ValueChange, element_factory):
    element = element_factory.lookup(change.element_id)
    return bool(
        element
        and getattr(element, change.property_name, None) != change.property_value
    )


@apply_change.register
def _(change: ValueChange, element_factory, modeling_factory):
    if change.applied:
        return
    element = element_factory[change.element_id]
    element.load(change.property_name, change.property_value)
    element.postload()
    change.applied = True


@applicable.register
def _(change: RefChange, element_factory):
    element = element_factory.lookup(change.element_id)
    ref = element_factory.lookup(change.property_ref)
    return bool(element and ref)


@apply_change.register
def _(change: RefChange, element_factory, modeling_factory):
    if change.applied:
        return
    element = element_factory[change.element_id]
    ref = element_factory[change.property_ref]
    prop = getattr(type(element), change.property_name, None)
    if change.op in ("add", "update"):
        element.load(change.property_name, ref)
        element.postload()
    elif change.op == "remove" and prop:
        if prop.upper == 1:
            delattr(element, change.property_name)
        else:
            getattr(element, change.property_name).remove(ref)
    change.applied = True

    if prop and prop.opposite:
        if other := next(
            element_factory.select(
                lambda e: isinstance(e, RefChange)
                and e.element_id == change.property_ref
                and e.property_ref == change.element_id
                and e.property_name == prop.opposite
            ),
            None,
        ):
            other.applied = True



"""Action diagram item."""

from gaphor import UML
from gaphor.diagram.presentation import ElementPresentation, Named, Valued
from gaphor.diagram.shapes import Box, Text, draw_border, stroke
from gaphor.diagram.support import represents
from gaphor.UML.recipes import stereotypes_str
from gaphor.UML.umlfmt import format_call_behavior_action_name


@represents(UML.Action)
class ActionItem(Named, ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=50, height=30)

        self.width = 100
        self.shape = Box(
            Text(
                text=lambda: stereotypes_str(self.subject),
            ),
            Text(
                text=lambda: self.subject.name or "",
            ),
            style={
                "padding": (4, 12, 4, 12),
                "border-radius": 15,
            },
            draw=draw_border,
        )

        self.watch("subject[NamedElement].name")
        self.watch("subject.appliedStereotype.classifier.name")


@represents(UML.ValueSpecificationAction)
class ValueSpecificationActionItem(Valued, ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=50, height=30)

        self.width = 100
        self.shape = Box(
            Text(
                text=lambda: "«valueSpecification»",
            ),
            Text(
                text=lambda: self.subject.value or "",
            ),
            style={
                "padding": (4, 12, 4, 12),
                "border-radius": 15,
            },
            draw=draw_border,
        )

        self.watch("subject[ValueSpecificationAction].value")


@represents(UML.CallBehaviorAction)
class CallBehaviorActionItem(ActionItem):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id)

        self.shape = Box(
            Text(
                text=lambda: stereotypes_str(self.subject),
            ),
            Text(
                text=lambda: format_call_behavior_action_name(self.subject),
            ),
            style={
                "padding": (4, 24, 4, 12),
                "border-radius": 15,
            },
            draw=self.draw_border_with_fork,
        )

        self.watch("subject[CallBehaviorAction].behavior.name")

    def draw_border_with_fork(self, box, context, bounding_box):
        draw_border(box, context, bounding_box)

        cr = context.cairo
        x, y, width, height = bounding_box

        x_offset = width - 15
        y_offset = height - 15
        half_width = 5
        half_height = 5

        cr.move_to(x_offset, y_offset - half_height)
        cr.line_to(x_offset, y_offset + half_height)
        cr.close_path()

        cr.move_to(x_offset - half_width, y_offset)
        cr.line_to(x_offset + half_width, y_offset)
        cr.close_path()

        cr.move_to(x_offset - half_width, y_offset)
        cr.line_to(x_offset - half_width, y_offset + half_height)
        cr.close_path()

        cr.move_to(x_offset + half_width, y_offset)
        cr.line_to(x_offset + half_width, y_offset + half_height)
        cr.close_path()

        stroke(context, fill=True)


@represents(UML.SendSignalAction)
class SendSignalActionItem(Named, ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=50, height=30)

        self.shape = Box(
            Text(
                text=lambda: stereotypes_str(self.subject),
            ),
            Text(text=lambda: self.subject.name or ""),
            style={"padding": (4, 24, 4, 12)},
            draw=self.draw_border,
        )

        self.watch("subject[NamedElement].name")
        self.watch("subject.appliedStereotype.classifier.name")

    def draw_border(self, box, context, bounding_box):
        cr = context.cairo
        d = 15
        x, y, width, height = bounding_box
        cr.move_to(0, 0)
        cr.line_to(width - d, 0)
        cr.line_to(width, height / 2)
        cr.line_to(width - d, height)
        cr.line_to(0, height)
        cr.close_path()

        stroke(context, fill=True)


@represents(UML.AcceptEventAction)
class AcceptEventActionItem(Named, ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=50, height=30)

        self.shape = Box(
            Text(
                text=lambda: stereotypes_str(self.subject),
            ),
            Text(text=lambda: self.subject.name or ""),
            style={"padding": (4, 12, 4, 24)},
            draw=self.draw_border,
        )

        self.watch("subject[NamedElement].name")
        self.watch("subject.appliedStereotype.classifier.name")

    def draw_border(self, box, context, bounding_box):
        cr = context.cairo
        d = 15
        x, y, width, height = bounding_box
        cr.move_to(0, 0)
        cr.line_to(width, 0)
        cr.line_to(width, height)
        cr.line_to(0, height)
        cr.line_to(d, height / 2)
        cr.close_path()

        stroke(context, fill=True)



"""Grouping functionality allows nesting of one item within another item
(parent item). This is useful in several use cases.

- artifact deployed within a node
- a class within a package, or a component
- composite structures (i.e. component within a node)
"""

from __future__ import annotations

import itertools
from typing import Callable

from generic.multidispatch import FunctionDispatcher, multidispatch

from gaphor.core.modeling import Diagram, Element, self_and_owners


def change_owner(new_parent, element):
    if new_parent and element.model is not new_parent.model:
        return False

    if new_parent and element.owner is new_parent:
        return False

    if new_parent is None and element.owner:
        return ungroup(element.owner, element)

    if not can_group(new_parent, element):
        return False

    if element.owner:
        ungroup(element.owner, element)

    return group(new_parent, element)


def no_group(parent, element) -> bool:
    return False


class GroupPreconditions:
    def __init__(self, func):
        self.__func = func

    def __getattr__(self, key):
        return getattr(self.__func, key)

    def __call__(self, parent, element) -> bool:
        if element in self_and_owners(parent):
            return False

        return self.__func(parent, element)  # type: ignore[no-any-return]


group: FunctionDispatcher[Callable[[Element, Element], bool]] = GroupPreconditions(
    multidispatch(object, object)(no_group)
)
group.register(None, object)(no_group)


def can_group(parent: Element, element_or_type: Element | type[Element]) -> bool:
    element_type = (
        type(element_or_type)
        if isinstance(element_or_type, Element)
        else element_or_type
    )
    parent_mro = type(parent).__mro__ if parent else [None]
    get_registration = group.registry.get_registration
    for t1, t2 in itertools.product(parent_mro, element_type.__mro__):
        if r := get_registration(t1, t2):
            return r is not no_group
    return False


@multidispatch(object, object)
def ungroup(parent, element) -> bool:
    return False


@ungroup.register(None, Element)
def none_ungroup(none, element):
    """In the rare (error?) case a model element has no parent, but is grouped
    in a diagram, allow it to ungroup."""
    return True


@group.register(Element, Diagram)
def diagram_group(element, diagram):
    diagram.element = element
    return True


@ungroup.register(Element, Diagram)
def diagram_ungroup(element, diagram):
    if diagram.element is element:
        del diagram.element
        return True
    return False



from __future__ import annotations

from pathlib import Path

from gaphor.abc import ActionProvider, Service
from gaphor.core import action, gettext
from gaphor.diagram.export import (
    escape_filename,
    save_eps,
    save_pdf,
    save_png,
    save_svg,
)
from gaphor.ui.filedialog import save_file_dialog


class DiagramExport(Service, ActionProvider):
    """Service for exporting diagrams as images (SVG, PNG, PDF)."""

    def __init__(self, diagrams=None, export_menu=None, main_window=None):
        self.diagrams = diagrams
        self.export_menu = export_menu
        self.main_window = main_window
        if export_menu:
            export_menu.add_actions(self)
        self.filename: Path = Path("export").absolute()

    def shutdown(self):
        if self.export_menu:
            self.export_menu.remove_actions(self)

    def save_dialog(self, diagram, title, ext, mime_type, handler):
        dot_ext = f".{ext}"
        filename = self.filename.with_name(
            escape_filename(diagram.name) or "export"
        ).with_suffix(dot_ext)

        def save_handler(filename):
            self.filename = filename
            handler(filename, diagram)

        save_file_dialog(
            title,
            save_handler,
            parent=self.main_window.window,
            filename=filename,
            extension=dot_ext,
            filters=[
                (gettext("All {ext} Files").format(ext=ext.upper()), dot_ext, mime_type)
            ],
        )

    @action(
        name="file-export-svg",
        label=gettext("Export as SVG"),
        tooltip=gettext("Export diagram as SVG"),
    )
    def save_svg_action(self):
        diagram = self.diagrams.get_current_diagram()
        self.save_dialog(
            diagram, gettext("Export diagram as SVG"), "svg", "image/svg+xml", save_svg
        )

    @action(
        name="file-export-png",
        label=gettext("Export as PNG"),
        tooltip=gettext("Export diagram as PNG"),
    )
    def save_png_action(self):
        diagram = self.diagrams.get_current_diagram()
        self.save_dialog(
            diagram, gettext("Export diagram as PNG"), "png", "image/png", save_png
        )

    @action(
        name="file-export-pdf",
        label=gettext("Export as PDF"),
        tooltip=gettext("Export diagram as PDF"),
    )
    def save_pdf_action(self):
        diagram = self.diagrams.get_current_diagram()
        self.save_dialog(
            diagram,
            gettext("Export diagram as PDF"),
            "pdf",
            "application/pdf",
            save_pdf,
        )

    @action(
        name="file-export-eps",
        label=gettext("Export as EPS"),
        tooltip=gettext("Export diagram as Encapsulated PostScript"),
    )
    def save_eps_action(self):
        diagram = self.diagrams.get_current_diagram()
        self.save_dialog(
            diagram,
            gettext("Export diagram as EPS"),
            "eps",
            "application/postscript",
            save_eps,
        )



"""Compiler for CSS selectors.

Based on cssselect2.compiler, written by Simon Sapin and Guillaume
Ayoub.
"""

import re
from functools import singledispatch
from typing import Callable, Dict, Iterator, Literal, Tuple, Union

import tinycss2

from gaphor.core.styling import selectors
from gaphor.core.styling.declarations import parse_declarations

# http://dev.w3.org/csswg/selectors/#whitespace
split_whitespace = re.compile("[^ \t\r\n\f]+").findall


Rule = Union[
    Tuple[Callable[[object], bool], Dict[str, object]],
    Tuple[Literal["error"], Union[tinycss2.ast.ParseError, selectors.SelectorError]],
]


def compile_style_sheet(*css: str) -> Iterator[Rule]:
    return (
        compiled_rule
        for _specificity, _order, compiled_rule in sorted(
            (
                ((-1,), order, (selspec, declarations))
                if selspec == "error"
                else (selspec[1], order, (selspec[0], declarations))
            )
            for order, (selspec, declarations) in enumerate(
                rule
                for sheet in css
                for rule in compile_rules(
                    tinycss2.parse_stylesheet(
                        sheet, skip_comments=True, skip_whitespace=True
                    )
                )
                if sheet
            )
        )
    )


def compile_rules(rules):
    for rule in rules:
        if rule.type == "error":
            yield "error", rule
            continue

        if rule.type == "at-rule" and rule.content:
            at_rules = tinycss2.parser.parse_rule_list(
                rule.content, skip_comments=True, skip_whitespace=True
            )
            media_selector = selectors.media_query_selector(rule.prelude)
            if not media_selector:
                continue
            media_query = compile_node(media_selector)
            yield from (
                ((_combine(media_query, selspec[0]), selspec[1]), declaration)
                for selspec, declaration in compile_rules(at_rules)
                if selspec != "error"
            )

        if rule.type != "qualified-rule":
            continue

        try:
            selector_list = compile_selector_list(rule.prelude)
        except selectors.SelectorError as e:
            yield "error", e
            continue

        declaration = {
            prop: value
            for prop, value in parse_declarations(rule.content)
            if prop != "error" and value is not None
        }

        yield from ((selector, declaration) for selector in selector_list)


def _combine(a, b):
    return lambda el: a(el) and b(el)


def compile_selector_list(input):
    """Compile a (comma-separated) list of selectors.

    Based on cssselect2.compiler.compile_selector_list().

    Returns a list of compiled selectors.
    """
    return [
        (compile_node(selector), selector.specificity)
        for selector in selectors.selectors(input)
    ]


@singledispatch
def compile_node(selector):
    """Dynamic dispatch selector nodes.

    Default behavior is to deny (no match).
    """
    raise selectors.SelectorError("Unknown selector", selector)


@compile_node.register
def compile_media_selector(selector: selectors.MediaSelector):
    query = selector.query
    if len(query) == 1:
        mode = query[0].lower()
    elif (
        len(query) == 3
        and query[0].lower() == "prefers-color-scheme"
        and query[1] == "="
    ):
        mode = query[2].lower()
    else:
        mode = None

    if mode in ("dark", "dark-mode"):
        return lambda el: el.dark_mode is True
    elif mode in ("light", "light-mode"):
        return lambda el: el.dark_mode is False
    return lambda el: False


@compile_node.register
def compile_compound_selector(selector: selectors.CompoundSelector):
    sub_expressions = [compile_node(sel) for sel in selector.simple_selectors]
    return lambda el: all(expr(el) for expr in sub_expressions)


@compile_node.register
def compile_name_selector(selector: selectors.LocalNameSelector):
    return lambda el: el.name() == selector.lower_local_name


def ancestors(el):
    if p := el.parent():
        yield p
        yield from ancestors(p)


def descendants(el):
    for c in el.children():
        yield c
        yield from descendants(c)


@compile_node.register
def compile_combined_selector(selector: selectors.CombinedSelector):
    left_inside = compile_node(selector.left)
    if selector.combinator == " ":

        def left(el):
            return any(left_inside(e) for e in ancestors(el))

    elif selector.combinator == ">":

        def left(el):
            p = el.parent()
            return p is not None and left_inside(p)

    else:
        raise selectors.SelectorError("Unknown combinator", selector.combinator)

    right = compile_node(selector.right)
    return lambda el: right(el) and left(el)


@compile_node.register
def compile_attribute_selector(selector: selectors.AttributeSelector):
    name = selector.lower_name
    operator = selector.operator
    value = selector.value and selector.value.lower()

    if operator is None:
        return lambda el: bool(el.attribute(name))
    elif operator == "=":
        return lambda el: el.attribute(name) == value
    elif operator == "~=":
        return lambda el: value in split_whitespace(el.attribute(name))
    elif operator == "^=":
        return lambda el: value and el.attribute(name).startswith(value)
    elif operator == "$=":
        return lambda el: value and el.attribute(name).endswith(value)
    elif operator == "*=":
        return lambda el: value and value in el.attribute(name)
    elif operator == "|=":

        def pipe_equal_matcher(el):
            v = el.attribute(name)
            return v == value or v and v.startswith(f"{value}-")

        return pipe_equal_matcher
    else:
        raise selectors.SelectorError("Unknown attribute operator", operator)


@compile_node.register
def compile_pseudo_class_selector(selector: selectors.PseudoClassSelector):
    name = selector.name
    if name == "empty":
        return lambda el: not next(el.children(), 0)
    elif name == "root":
        return lambda el: not el.parent()
    elif name in ("hover", "focus", "active", "drop", "disabled"):
        return lambda el: name in el.state()
    else:
        raise selectors.SelectorError("Unknown pseudo-class", name)


@compile_node.register
def compile_functional_pseudo_class_selector(
    selector: selectors.FunctionalPseudoClassSelector,
):
    name = selector.name
    if name not in ("has", "is", "not"):
        raise selectors.SelectorError("Unknown pseudo-class", name)

    sub_selectors = compile_selector_list(selector.arguments)
    selector.specificity = max(spec for _, spec in sub_selectors)
    if name == "has":
        return lambda el: any(
            any(sel(c) for sel, _ in sub_selectors) for c in descendants(el)
        )
    elif name == "is":
        return lambda el: any(sel(el) for sel, _ in sub_selectors)
    elif name == "not":
        return lambda el: not any(sel(el) for sel, _ in sub_selectors)


@compile_node.register
def compile_pseudo_element_selector(selector: selectors.PseudoElementSelector):
    name = selector.name
    if name != "after":
        raise selectors.SelectorError("Unknown pseudo-element", name)

    return lambda el: name == el.pseudo



"""Defines a status window class for displaying the progress of a queue."""


from gaphas.decorators import g_async
from gi.repository import Gtk, Pango


class StatusWindow:
    """Create a borderless window on the parent, usually the main window, with
    a label and a progress bar.

    The progress bar is updated as the queue is updated.
    """

    def __init__(self, title, message, parent=None):
        """Create the status window.

        The title parameter is the title of the window.  The message
        parameter is a string displayed near the progress bar to
        indicate what is happening.  The parent parameter is the parent
        window to display the window in.
        """

        self.title = title
        self.message = message
        self.parent = parent
        self.window: Gtk.Window = None

        self.display()

    def init_window(self):
        frame = Gtk.Frame.new(None)
        vbox = Gtk.Box.new(Gtk.Orientation.VERTICAL, spacing=12)
        label = Gtk.Label.new(self.message)
        label.set_ellipsize(Pango.EllipsizeMode.MIDDLE)

        self.progress_bar = Gtk.ProgressBar.new()
        self.progress_bar.set_size_request(400, -1)

        self.window = Gtk.Window.new()
        self.window.set_child(frame)
        frame.set_child(vbox)
        vbox.append(label)
        vbox.append(self.progress_bar)

        self.window.set_title(self.title)
        self.window.add_css_class("status-window")
        if self.parent:
            self.window.set_transient_for(self.parent)
        self.window.set_modal(True)
        self.window.set_resizable(False)
        self.window.set_decorated(False)

    @g_async()
    def display(self):
        if not self.window:
            self.init_window()

        assert self.window

        self.window.set_visible(True)

    def progress(self, percentage: int):
        """Update progress percentage (0..100)."""
        if self.progress_bar:
            self.progress_bar.set_fraction(min(percentage, 100.0) / 100.0)

    def destroy(self):
        """Destroy the status window.

        This will also remove the gobject handler.
        """
        if self.window:
            self.window.destroy()
            self.window = None



from __future__ import annotations

import logging

from gi.repository import Gio, GObject, Gtk

from gaphor import UML
from gaphor.core import gettext, transactional
from gaphor.core.format import format, parse
from gaphor.diagram.propertypages import (
    NamePropertyPage,
    PropertyPageBase,
    PropertyPages,
    help_link,
    new_resource_builder,
    unsubscribe_all_on_destroy,
)
from gaphor.UML.classes.datatype import DataTypeItem
from gaphor.UML.classes.interface import Folded, InterfaceItem
from gaphor.UML.classes.klass import ClassItem
from gaphor.UML.deployments.connector import ConnectorItem
from gaphor.UML.propertypages import (
    check_button_handlers,
    create_list_store,
    list_item_factory,
    list_view_activated,
    list_view_key_handler,
    text_field_handlers,
    update_list_store,
)

log = logging.getLogger(__name__)


new_builder = new_resource_builder("gaphor.UML.classes")


@PropertyPages.register(UML.NamedElement)
class NamedElementPropertyPage(NamePropertyPage):
    """An adapter which works for any named item view.

    It also sets up a table view which can be extended.
    """

    order = 10

    def __init__(self, subject: UML.NamedElement | None):
        super().__init__(subject)

    def construct(self):
        if (
            not self.subject
            or UML.recipes.is_metaclass(self.subject)
            or isinstance(
                self.subject, (UML.ActivityPartition, UML.ActivityParameterNode)
            )
        ):
            return

        return super().construct()


@PropertyPages.register(UML.Classifier)
class ClassifierPropertyPage(PropertyPageBase):
    order = 15

    def __init__(self, subject):
        self.subject = subject

    def construct(self):
        if UML.recipes.is_metaclass(self.subject):
            return

        builder = new_builder(
            "classifier-editor",
            signals={"abstract-changed": (self._on_abstract_change,)},
        )

        abstract = builder.get_object("abstract")
        abstract.set_active(self.subject.isAbstract)

        return builder.get_object("classifier-editor")

    @transactional
    def _on_abstract_change(self, button, gparam):
        self.subject.isAbstract = button.get_active()


@PropertyPages.register(InterfaceItem)
class InterfacePropertyPage(PropertyPageBase):
    """Adapter which shows a property page for an interface view."""

    order = 15

    def __init__(self, item):
        self.item = item

    def construct(self):
        builder = new_builder(
            "interface-editor", signals={"folded-changed": (self._on_fold_change,)}
        )

        item = self.item

        connected_items = [
            c.item for c in item.diagram.connections.get_connections(connected=item)
        ]
        disallowed = (ConnectorItem,)
        can_fold = not any(isinstance(i, disallowed) for i in connected_items)

        folded = builder.get_object("folded")
        folded.set_active(item.folded != Folded.NONE)
        folded.set_sensitive(can_fold)

        return builder.get_object("interface-editor")

    @transactional
    def _on_fold_change(self, button, gparam):
        item = self.item

        fold = button.get_active()

        item.folded = Folded.PROVIDED if fold else Folded.NONE


class AttributeView(GObject.Object):
    def __init__(self, attr: UML.Property | None, klass: UML.Class):
        super().__init__()
        self.attr = attr
        self.klass = klass

    @GObject.Property(type=str, default="")
    def attribute(self):
        return format(self.attr, note=True) if self.attr else ""

    @attribute.setter  # type: ignore[no-redef]
    @transactional
    def attribute(self, value):
        if not self.attr:
            if not value:
                return

            model = self.klass.model
            self.attr = model.create(UML.Property)
            self.klass.ownedAttribute = self.attr
        parse(self.attr, value)

    @GObject.Property(type=bool, default=False)
    def static(self):
        return self.attr.isStatic if self.attr else False

    @static.setter  # type: ignore[no-redef]
    @transactional
    def static(self, value):
        if not self.attr:
            return

        self.attr.isStatic = value

    editing = GObject.Property(type=bool, default=False)

    def start_editing(self):
        self.editing = True

    def empty(self):
        return not self.attr

    def unlink(self):
        if self.attr:
            self.attr.unlink()

    def swap(self, item1, item2):
        return self.klass.ownedAttribute.swap(item1.attr, item2.attr)


def attribute_model(klass: UML.Class) -> Gio.ListStore:
    return create_list_store(
        AttributeView,
        (a for a in klass.ownedAttribute if not a.association),
        lambda attr: AttributeView(attr, klass),
    )


def update_attribute_model(store: Gio.ListStore, klass: UML.Class) -> None:
    update_list_store(
        store,
        lambda item: item.attr,
        (a for a in klass.ownedAttribute if not a.association),
        lambda attr: AttributeView(attr, klass),
    )


@PropertyPages.register(DataTypeItem)
@PropertyPages.register(ClassItem)
@PropertyPages.register(InterfaceItem)
class AttributesPage(PropertyPageBase):
    """An editor for attributes associated with classes and interfaces."""

    order = 20

    def __init__(self, item):
        super().__init__()
        self.item = item
        self.watcher = item.subject and item.subject.watcher()

    def construct(self):
        if not self.item.subject:
            return

        builder = new_builder(
            "attributes-editor",
            "attributes-info",
            signals={
                "attributes-activated": (list_view_activated,),
                "attributes-key-pressed": (list_view_key_handler,),
                "show-attributes-changed": (self.on_show_attributes_changed,),
                "attributes-info-clicked": (self.on_attributes_info_clicked,),
            },
        )
        self.info = builder.get_object("attributes-info")
        help_link(builder, "attributes-info-icon", "attributes-info")

        show_attributes = builder.get_object("show-attributes")
        show_attributes.set_active(self.item.show_attributes)

        column_view: Gtk.ColumnView = builder.get_object("attributes-list")

        for column, factory in zip(
            column_view.get_columns(),
            [
                list_item_factory(
                    "text-field-cell.ui",
                    klass=AttributeView,
                    attribute=AttributeView.attribute,
                    placeholder_text=gettext("New Attribute…"),
                    signal_handlers=text_field_handlers("attribute"),
                ),
                list_item_factory(
                    "check-button-cell.ui",
                    klass=AttributeView,
                    attribute=AttributeView.static,
                    signal_handlers=check_button_handlers("static"),
                ),
            ],
        ):
            column.set_factory(factory)

        self.model = attribute_model(self.item.subject)
        selection = Gtk.SingleSelection.new(self.model)
        column_view.set_model(selection)

        if self.watcher:
            self.watcher.watch("ownedAttribute", self.on_attributes_changed)

        return unsubscribe_all_on_destroy(
            builder.get_object("attributes-editor"), self.watcher
        )

    @transactional
    def on_show_attributes_changed(self, button, gparam):
        self.item.show_attributes = button.get_active()
        self.item.request_update()

    def on_attributes_changed(self, event):
        update_attribute_model(self.model, self.item.subject)

    def on_attributes_info_clicked(self, image, event):
        self.info.set_visible(True)


class OperationView(GObject.Object):
    def __init__(self, oper: UML.Operation | None, klass: UML.Class):
        super().__init__()
        self.oper = oper
        self.klass = klass

    @GObject.Property(type=str, default="")
    def operation(self):
        return format(self.oper, note=True) if self.oper else ""

    @operation.setter  # type: ignore[no-redef]
    @transactional
    def operation(self, value):
        if not self.oper:
            if not value:
                return

            model = self.klass.model
            self.oper = model.create(UML.Operation)
            self.klass.ownedOperation = self.oper
        parse(self.oper, value)

    @GObject.Property(type=bool, default=False)
    def static(self):
        return self.oper.isStatic if self.oper else False

    @static.setter  # type: ignore[no-redef]
    @transactional
    def static(self, value):
        if not self.oper:
            return

        self.oper.isStatic = value

    @GObject.Property(type=bool, default=False)
    def abstract(self):
        return self.oper.isAbstract if self.oper else False

    @abstract.setter  # type: ignore[no-redef]
    @transactional
    def abstract(self, value):
        if not self.oper:
            return

        self.oper.isAbstract = value

    editing = GObject.Property(type=bool, default=False)

    def start_editing(self):
        self.editing = True

    def empty(self):
        return not self.oper

    def unlink(self):
        if self.oper:
            self.oper.unlink()

    def swap(self, item1, item2):
        return self.klass.ownedOperation.swap(item1.oper, item2.oper)


def operation_model(klass: UML.Class) -> Gio.ListStore:
    return create_list_store(
        OperationView,
        klass.ownedOperation,
        lambda oper: OperationView(oper, klass),
    )


def update_operation_model(store: Gio.ListStore, klass: UML.Class) -> None:
    update_list_store(
        store,
        lambda item: item.oper,
        klass.ownedOperation,
        lambda oper: OperationView(oper, klass),
    )


@PropertyPages.register(DataTypeItem)
@PropertyPages.register(ClassItem)
@PropertyPages.register(InterfaceItem)
class OperationsPage(PropertyPageBase):
    """An editor for operations associated with classes and interfaces."""

    order = 30

    def __init__(self, item):
        super().__init__()
        self.item = item
        self.watcher = item.subject and item.subject.watcher()

    def construct(self):
        if not self.item.subject:
            return

        builder = new_builder(
            "operations-editor",
            "operations-info",
            signals={
                "operations-activated": (list_view_activated,),
                "operations-key-pressed": (list_view_key_handler,),
                "show-operations-changed": (self.on_show_operations_changed,),
                "operations-info-clicked": (self.on_operations_info_clicked,),
            },
        )

        self.info = builder.get_object("operations-info")
        help_link(builder, "operations-info-icon", "operations-info")

        show_operations = builder.get_object("show-operations")
        show_operations.set_active(self.item.show_operations)

        column_view: Gtk.ColumnView = builder.get_object("operations-list")

        for column, factory in zip(
            column_view.get_columns(),
            [
                list_item_factory(
                    "text-field-cell.ui",
                    klass=OperationView,
                    attribute=OperationView.operation,
                    placeholder_text=gettext("New Operation…"),
                    signal_handlers=text_field_handlers("operation"),
                ),
                list_item_factory(
                    "check-button-cell.ui",
                    klass=OperationView,
                    attribute=OperationView.abstract,
                    signal_handlers=check_button_handlers("abstract"),
                ),
                list_item_factory(
                    "check-button-cell.ui",
                    klass=OperationView,
                    attribute=OperationView.static,
                    signal_handlers=check_button_handlers("static"),
                ),
            ],
        ):
            column.set_factory(factory)

        self.model = operation_model(self.item.subject)
        selection = Gtk.SingleSelection.new(self.model)
        column_view.set_model(selection)

        if self.watcher:
            self.watcher.watch("ownedOperation", self.on_operations_changed)

        return unsubscribe_all_on_destroy(
            builder.get_object("operations-editor"), self.watcher
        )

    @transactional
    def on_show_operations_changed(self, button, gparam):
        self.item.show_operations = button.get_active()
        self.item.request_update()

    def on_operations_changed(self, event):
        update_operation_model(self.model, self.item.subject)

    def on_operations_info_clicked(self, image, event):
        self.info.set_visible(True)


@PropertyPages.register(UML.Component)
class ComponentPropertyPage(PropertyPageBase):
    order = 15

    subject: UML.Component

    def __init__(self, subject):
        self.subject = subject

    def construct(self):
        subject = self.subject

        if not subject:
            return

        builder = new_builder(
            "component-editor",
            signals={"indirectly-instantiated-changed": (self._on_ii_change,)},
        )

        ii = builder.get_object("indirectly-instantiated")
        ii.set_active(subject.isIndirectlyInstantiated)

        return builder.get_object("component-editor")

    @transactional
    def _on_ii_change(self, button, gparam):
        """Called when user clicks "Indirectly instantiated" check button."""
        if subject := self.subject:
            subject.isIndirectlyInstantiated = button.get_active()



"""
Plugins
=======

This module contains a bunch of standard plugins.
Plugins can be registered in Gaphor by declaring them as service entry point::

    entry_points = {
        'gaphor.services': [
            'xmi_export = gaphor.plugins.xmiexport:XMIExport',
        ],
    },


There is a thin line between a service and a plugin. A service typically performs some basic need for the applications (such as the element factory or the undo mechanism). A plugin is more of an add-on. For example a plugin can depend on external libraries or provide a cross-over function between two applications.

"""
import contextlib
import importlib.abc
import importlib.machinery
import importlib.metadata
import os
import pathlib
import sys

from gaphor import entrypoint

PLUGIN_VERSION = 2


def default_plugin_path() -> pathlib.Path:
    if plugin_path_envvar := os.getenv("GAPHOR_PLUGIN_PATH"):
        return pathlib.Path(plugin_path_envvar)
    return pathlib.Path.home() / ".local" / "gaphor" / f"plugins-{PLUGIN_VERSION}"


@contextlib.contextmanager
def enable_plugins(plugin_path: pathlib.Path):
    entrypoint.list_entry_points.cache_clear()
    path_finder = PluginMetaPathFinder(plugin_path)
    sys.meta_path.append(path_finder)
    yield
    sys.meta_path.remove(path_finder)


class PluginMetaPathFinder(importlib.abc.MetaPathFinder):
    def __init__(self, plugin_path: pathlib.Path):
        self.plugin_path = str(plugin_path)

    def find_spec(self, fullname, path=None, target=None):
        return importlib.machinery.PathFinder.find_spec(
            fullname, path=[self.plugin_path], target=target
        )

    def find_distributions(self, *args, **kwargs):
        return importlib.metadata.MetadataPathFinder.find_distributions(
            context=importlib.metadata.DistributionFinder.Context(
                path=[self.plugin_path]
            )
        )



"""For ease of creation, maintain a mapping from Element to Diagram Item."""

from typing import Dict

from gaphor.core.modeling import Element, Presentation


def represents(uml_element, **metadata):
    """A decorator to assign a default Element type to a diagram item."""

    def wrapper(presentation):
        set_diagram_item(uml_element, presentation, metadata)
        return presentation

    return wrapper


# Map elements to their (default) representation.
_element_to_item_map: Dict[Element, Presentation] = {}
_item_to_metadata_map: Dict[Presentation, dict[str, object]] = {}


def get_diagram_item(element_cls):
    global _element_to_item_map
    return _element_to_item_map.get(element_cls)


def get_diagram_item_metadata(item_cls):
    global _item_to_metadata_map
    return _item_to_metadata_map.get(item_cls, {})


def get_model_element(item_cls):
    global _element_to_item_map
    elements = [
        element
        for element, presentation in _element_to_item_map.items()
        if item_cls is presentation
    ]
    return elements[0] if elements else None


def set_diagram_item(element, item, metadata):
    assert element not in _element_to_item_map
    _element_to_item_map[element] = item
    _item_to_metadata_map[item] = metadata



from gi.repository import Adw

from gaphor.core import event_handler
from gaphor.event import Notification


class InAppNotifier:
    def __init__(self, toast_overlay):
        self.toast_overlay = toast_overlay
        self.latest_message: str | None = None

    @event_handler(Notification)
    def handle(self, event: Notification):
        if event.message != self.latest_message:
            self.latest_message = event.message
            toast = Adw.Toast.new(event.message)
            toast.connect("dismissed", self._on_dismissed)
            self.toast_overlay.add_toast(toast)
            return toast
        return None

    def _on_dismissed(self, toast):
        self.latest_message = None



"""The Toolbox which is the tool palette.

This is the toolbox in the lower left of the screen.
"""

import functools
import logging
from typing import Optional, Tuple

from gi.repository import Gdk, GLib, GObject, Gtk

from gaphor.core.eventmanager import EventManager, event_handler
from gaphor.diagram.diagramtoolbox import ToolboxDefinition
from gaphor.diagram.event import ToolCompleted
from gaphor.diagram.tools.dnd import ToolboxActionDragData
from gaphor.services.modelinglanguage import (
    ModelingLanguageChanged,
    ModelingLanguageService,
)
from gaphor.services.properties import Properties
from gaphor.ui.abc import UIComponent
from gaphor.ui.event import CurrentDiagramChanged, ToolSelected

log = logging.getLogger(__name__)


class Toolbox(UIComponent):
    def __init__(
        self,
        event_manager: EventManager,
        properties: Properties,
        modeling_language: ModelingLanguageService,
    ):
        self.event_manager = event_manager
        self.properties = properties
        self.modeling_language = modeling_language
        self._toolbox: Optional[Gtk.Box] = None
        self._toolbox_container: Optional[Gtk.ScrolledWindow] = None
        self._current_diagram_type = ""

    def open(self) -> Gtk.ScrolledWindow:
        self._toolbox = self.create_toolbox(self.modeling_language.toolbox_definition)
        self._toolbox_container = create_toolbox_container(self._toolbox)
        self.event_manager.subscribe(self._on_diagram_item_placed)
        self.event_manager.subscribe(self._on_modeling_language_changed)
        self.event_manager.subscribe(self._on_current_diagram_changed)
        self.select_tool("toolbox-pointer")
        return self._toolbox_container

    def close(self) -> None:
        if self._toolbox:
            if self._toolbox_container:
                self._toolbox_container.unparent()
            self._toolbox = None
        self.event_manager.unsubscribe(self._on_modeling_language_changed)
        self.event_manager.unsubscribe(self._on_diagram_item_placed)
        self.event_manager.unsubscribe(self._on_current_diagram_changed)

    def activate_shortcut(self, keyval: int, state: Gdk.ModifierType) -> bool:
        # Accelerator keys are lower case. Since we handle them in a key-press event
        # handler, we'll need the upper-case versions as well in case Shift is pressed.
        state = state & Gtk.accelerator_get_default_mod_mask()
        for _title, items in self.modeling_language.toolbox_definition:
            for action_name, _label, _icon_name, shortcut, *_rest in items:
                if not shortcut:
                    continue
                keys, mod = parse_shortcut(shortcut)
                if state == mod and keyval in keys:
                    self.select_tool(action_name)
                    return True
        return False

    @property
    def active_tool_name(self):
        for flowbox in self.flowboxes():
            if children := flowbox.get_selected_children():
                return children[0].action_name
        return "toolbox-pointer"

    def create_toolbox(self, toolbox_actions: ToolboxDefinition) -> Gtk.Box:
        toolbox = Gtk.Box.new(Gtk.Orientation.VERTICAL, 0)
        expanded = self.expanded_sections()

        def on_expanded(widget, _prop, index, revealer):
            expanded = widget.get_property("expanded")
            self.expanded_sections(index, expanded)
            revealer.set_reveal_child(expanded)

        for index, (title, items) in enumerate(toolbox_actions):
            revealer = Gtk.Revealer.new()
            revealer.set_transition_type(Gtk.RevealerTransitionType.SLIDE_UP)
            revealer.set_reveal_child(expanded[index])
            expander = Gtk.Expander.new(title)
            expander.set_property("expanded", expanded[index])
            expander.connect("notify::expanded", on_expanded, index, revealer)
            flowbox = Gtk.FlowBox.new()
            flowbox.set_homogeneous(True)
            flowbox.set_row_spacing(1)
            flowbox.set_column_spacing(1)
            flowbox.set_max_children_per_line(12)
            flowbox.connect("child-activated", self._on_tool_activated)

            drag_source = Gtk.DragSource.new()
            drag_source.connect("prepare", _flowbox_drag_prepare)
            flowbox.add_controller(drag_source)

            for action_name, label, icon_name, shortcut, item_factory, *_rest in items:
                button = create_toolbox_button(
                    action_name, icon_name, label, shortcut, bool(item_factory)
                )
                flowbox.insert(button, -1)

            revealer.set_child(flowbox)
            toolbox.append(expander)
            toolbox.append(revealer)

        return toolbox

    def expanded_sections(self, index=None, state=None):
        diagram_type = self._current_diagram_type or ""
        expanded_property = f"toolbox-{self.modeling_language.active_modeling_language}-{diagram_type}-expanded"
        expanded_sections = next(
            (
                diagram.sections
                for diagram in self.modeling_language.diagram_types
                if diagram.id == diagram_type
            ),
            (),
        )
        default_expanded = {
            i: i == 0 or not bool(expanded_sections)
            for i, _ in enumerate(self.modeling_language.toolbox_definition)
        }
        expanded = default_expanded | self.properties.get(expanded_property, {})

        if index is not None:
            expanded[index] = state
            self.properties.set(expanded_property, expanded)

        required_expanded = {
            i: True
            for i, section in enumerate(self.modeling_language.toolbox_definition)
            if section in expanded_sections
        }
        return expanded | required_expanded

    def flowboxes(self):
        if self._toolbox:
            for expander in iter_children(self._toolbox):
                if isinstance(expander, Gtk.Revealer):
                    yield expander.get_child()

    def select_tool(self, tool_name: str) -> None:
        for flowbox in self.flowboxes():
            flowbox.unselect_all()
            for child in iter_children(flowbox):
                if child.action_name == tool_name:
                    flowbox.select_child(child)
                    self.event_manager.handle(ToolSelected(tool_name))

    @event_handler(ToolCompleted)
    def _on_diagram_item_placed(self, event) -> None:
        if self.properties.get("reset-tool-after-create", True):
            # Select tool from an idle handler, so the original tool can complete properly.
            GLib.idle_add(self.select_tool, "toolbox-pointer")

    @event_handler(ModelingLanguageChanged)
    def _on_modeling_language_changed(self, event: ModelingLanguageChanged) -> None:
        """Reconfigures the toolbox based on the modeling language selected."""
        toolbox = self.create_toolbox(self.modeling_language.toolbox_definition)
        if self._toolbox_container:
            self._toolbox_container.set_child(toolbox)
        self._toolbox = toolbox
        self.select_tool("toolbox-pointer")

    @event_handler(CurrentDiagramChanged)
    def _on_current_diagram_changed(self, event: CurrentDiagramChanged) -> None:
        if not self._toolbox:
            return

        self._current_diagram_type = event.diagram and event.diagram.diagramType or ""

        expanded = self.expanded_sections()
        for index, expander in enumerate(
            c for c in iter_children(self._toolbox) if isinstance(c, Gtk.Expander)
        ):
            expander.set_property("expanded", expanded[index])

    def _on_tool_activated(self, flowbox, child):
        self.select_tool(child.action_name)


def create_toolbox_container(toolbox: Gtk.Widget) -> Gtk.ScrolledWindow:
    toolbox_container = Gtk.ScrolledWindow()
    toolbox_container.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
    toolbox_container.set_child(toolbox)
    return toolbox_container


def create_toolbox_button(
    action_name: str,
    icon_name: str,
    label: str,
    shortcut: Optional[str],
    draggable: bool,
) -> Gtk.Button:
    """Creates a tool button for the toolbox."""
    button = Gtk.FlowBoxChild.new()
    icon = Gtk.Image.new_from_icon_name(icon_name)
    button.set_child(icon)
    button.action_name = action_name
    button.icon_name = icon_name
    button.draggable = draggable
    if label:
        if shortcut:
            _, a, m = Gtk.accelerator_parse(shortcut)
            button.set_tooltip_text(f"{label} ({Gtk.accelerator_get_label(a, m)})")
        else:
            button.set_tooltip_text(f"{label}")

    return button


def iter_children(widget):
    child = widget.get_first_child()
    while child:
        yield child
        child = child.get_next_sibling()


def _flowbox_drag_prepare(source: Gtk.DragSource, x: int, y: int):
    child = source.get_widget().get_child_at_pos(x, y)
    if not child.draggable:
        return None

    display = Gdk.Display.get_default()
    theme_icon = Gtk.IconTheme.get_for_display(display).lookup_icon(
        child.icon_name,
        None,
        24,
        1,
        Gtk.TextDirection.NONE,
        Gtk.IconLookupFlags.FORCE_SYMBOLIC,
    )
    source.set_icon(theme_icon, 0, 0)

    v = GObject.Value(
        ToolboxActionDragData.__gtype__,
        ToolboxActionDragData(action=child.action_name),
    )
    return Gdk.ContentProvider.new_for_value(v)


_upper_offset: int = ord("A") - ord("a")


@functools.cache
def parse_shortcut(shortcut: str) -> Tuple[Tuple[int, int], Gdk.ModifierType]:
    _, key, mod = Gtk.accelerator_parse(shortcut)
    return (key, key + _upper_offset), mod



from gaphor.diagram.group import group, no_group
from gaphor.UML.uml import Interaction, Lifeline, Message, Type


@group.register(Interaction, Lifeline)
def interaction_lifeline_group(parent, element):
    """Add lifeline to interaction."""
    element.interaction = parent
    return True


@group.register(Interaction, Message)
def interaction_message_group(parent, element):
    """Add message to interaction."""
    element.interaction = parent
    return True


group.register(Interaction, Type)(no_group)



from __future__ import annotations

from typing import Iterable

from gi.repository import Gtk

from gaphor.core.eventmanager import EventManager
from gaphor.transaction import Transaction


class TxData:
    def __init__(self, event_manager):
        self.event_manager = event_manager
        self.txs: list[Transaction] = []

    def begin(self):
        self.txs.append(Transaction(self.event_manager))

    def commit(self):
        assert self.txs
        tx = self.txs.pop()
        tx.commit()


def transactional_tool(
    *tools: Gtk.Gesture, event_manager: EventManager | None = None
) -> Iterable[Gtk.Gesture]:
    tx_data = TxData(event_manager)
    for tool in tools:
        tool.connect("begin", on_begin, tx_data)
        tool.connect_after("end", on_end, tx_data)
    return tools


def on_begin(gesture, _sequence, tx_data):
    tx_data.begin()


def on_end(gesture, _sequence, tx_data):
    tx_data.commit()



from gaphor.core.styling import JustifyContent, TextAlign
from gaphor.diagram.shapes import BoundedBox, Text, draw_top_separator


def region_compartment(subject):
    if not subject:
        return

    for index, region in enumerate(subject.region):
        yield _create_region_compartment(region, index)


def _create_region_compartment(region, index):
    return BoundedBox(
        Text(
            text=lambda: region.name or "",
            style={"text-align": TextAlign.LEFT},
        ),
        style={
            "padding": (4, 4, 4, 4),
            "min-height": 100,
            "justify-content": JustifyContent.START,
            "dash-style": (7, 3) if index > 0 else (),
        },
        draw=draw_top_separator,
    )
